# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""ExplainManager."""

import os
import threading
import time
from collections import OrderedDict
from datetime import datetime
from typing import Optional

from mindinsight.conf import settings
from mindinsight.datavisual.common import exceptions
from mindinsight.datavisual.common.enums import BaseEnum
from mindinsight.datavisual.data_access.file_handler import FileHandler
from mindinsight.datavisual.data_transform.summary_watcher import SummaryWatcher
from mindinsight.explainer.common.log import logger
from mindinsight.explainer.manager.explain_loader import ExplainLoader, _LoaderStatus
from mindinsight.utils.exceptions import ParamValueError, UnknownError

_MAX_LOADERS_NUM = 3


class _ExplainManagerStatus(BaseEnum):
    """Manager status."""
    INIT = 'INIT'
    LOADING = 'LOADING'
    STOPPING = 'STOPPING'
    DONE = 'DONE'


class ExplainManager:
    """ExplainManager."""

    def __init__(self, summary_base_dir: str):
        self._summary_base_dir = summary_base_dir
        self._loader_pool = OrderedDict()
        self._loading_status = _ExplainManagerStatus.INIT.value
        self._status_mutex = threading.Lock()
        self._load_data_mutex = threading.Lock()
        self._loader_pool_mutex = threading.Lock()
        self._max_loaders_num = _MAX_LOADERS_NUM
        self._summary_watcher = SummaryWatcher()

    @property
    def summary_base_dir(self):
        """Return the base directory for summary records."""
        return self._summary_base_dir

    def start_load_data(self, reload_interval: int = 0):
        """
        Start individual thread to cache explain_jobs and loading summary data periodically.

        Args:
            reload_interval (int): Specify the loading period in seconds. If interval == 0, data will only be loaded
                once. Default: 0.
        """
        thread = threading.Thread(target=self._repeat_loading,
                                  name='explainer.start_load_thread',
                                  args=(reload_interval,),
                                  daemon=True)
        thread.start()

    def get_job(self, loader_id: str) -> Optional[ExplainLoader]:
        """
        Return ExplainLoader given loader_id.

        If explain job w.r.t given loader_id is not found, None will be returned.

        Args:
            loader_id (str): The id of expected ExplainLoader.

        Returns:
            ExplainLoader, the data loader specified by loader_id.
        """
        self._check_status_valid()

        with self._loader_pool_mutex:
            if loader_id in self._loader_pool:
                self._loader_pool[loader_id].query_time = datetime.now().timestamp()
                self._loader_pool.move_to_end(loader_id, last=True)
                loader = self._loader_pool[loader_id]
                if loader.status == _LoaderStatus.STOP.value:
                    self._reload_data_again()
                return self._loader_pool[loader_id]

            try:
                loader = self._generate_loader_from_relative_path(loader_id)
                loader.query_time = datetime.now().timestamp()
                self._add_loader(loader)
                self._reload_data_again()
            except ParamValueError:
                logger.warning('Cannot find summary in path: %s. No explain_job will be returned.', loader_id)
                return None
        return loader

    def get_job_list(self, offset=0, limit=None):
        """
        Return List of explain jobs. includes job ID, create and update time.

        Args:
            offset (int): An offset for page. Ex, offset is 0, mean current page is 1. Default: 0.
            limit (int): The max data items for per page. Default: 10.

        Returns:
            tuple, the elements of the returned tuple are:

                - total (int): The overall number of explain directories
                - dir_infos (list): List of summary directory info including the following attributes:

                    - relative_path (str): Relative path of summary directory, referring to settings.SUMMARY_BASE_DIR,
                         starting with "./".
                    - create_time (datetime): Creation time of summary file.
                    - update_time (datetime): Modification time of summary file.
        """
        total, dir_infos = \
            self._summary_watcher.list_explain_directories(self._summary_base_dir, offset=offset, limit=limit)
        return total, dir_infos

    def _repeat_loading(self, repeat_interval):
        """Periodically loading summary."""
        # Allocate CPU resources to enable gunicorn to start the web service.
        time.sleep(1)
        while True:
            try:
                if self.status == _ExplainManagerStatus.STOPPING.value:
                    logger.debug('Current loading status is %s, we will not trigger repeat loading.',
                                 _ExplainManagerStatus.STOPPING.value)
                else:
                    logger.info('Starts triggering repeat loading, repeat interval: %r.', repeat_interval)
                    self._load_data()
                    if not repeat_interval:
                        return
                time.sleep(repeat_interval)
            except UnknownError as ex:
                logger.error('Unexpected error happens when loading data. Loading status: %s, loading pool size: %d'
                             'Detail: %s', self.status, len(self._loader_pool), str(ex))

    def _load_data(self):
        """
        Prepare loaders in cache and start loading the data from summaries.

        Only a limited number of loaders will be cached in terms of updated_time or query_time. The size of cache
        pool is determined by _MAX_LOADERS_NUM. When the manager start loading data, only the latest _MAX_LOADER_NUM
        summaries will be loaded in cache. If a cached loader if queries by 'get_job', the query_time of the loader
        will be updated as well as the the loader moved to the end of cache. If an uncached summary is queried,
        a new loader instance will be generated and put to the end cache.
        """
        try:
            with self._load_data_mutex:
                if self.status == _ExplainManagerStatus.LOADING.value:
                    logger.info('Current status is %s, will ignore to load data.', self.status)
                    return

                logger.info('Start to load data, and status change to %s.', _ExplainManagerStatus.LOADING.value)
                self.status = _ExplainManagerStatus.LOADING.value
                self._cache_loaders()

                if self.status == _ExplainManagerStatus.STOPPING.value:
                    logger.info('The manager status has been %s, will not execute loading.', self.status)
                    return
                self._execute_loading()

                logger.info('Load event data end, current status: %s, next status: %s, loader pool size: %d.',
                            self.status, _ExplainManagerStatus.DONE.value, len(self._loader_pool))

        except Exception as ex:
            logger.exception(ex)
            raise UnknownError(str(ex))
        finally:
            self.status = _ExplainManagerStatus.DONE.value

    def _cache_loaders(self):
        """Cache explain loader in cache pool."""
        dir_map_mtimes = []
        _, summaries_info = self._summary_watcher.list_explain_directories(self._summary_base_dir)

        for summary_info in summaries_info:
            summary_path = summary_info.get('relative_path')
            summary_update_time = summary_info.get('update_time').timestamp()

            if summary_path in self._loader_pool:
                summary_update_time = max(summary_update_time, self._loader_pool[summary_path].query_time)

            dir_map_mtimes.append((summary_info, summary_update_time))

        sorted_summaries_info = sorted(dir_map_mtimes, key=lambda x: x[1])[-_MAX_LOADERS_NUM:]

        with self._loader_pool_mutex:
            for summary_info, query_time in sorted_summaries_info:
                summary_path = summary_info['relative_path']
                if summary_path not in self._loader_pool:
                    loader = self._generate_loader_from_relative_path(summary_path)
                    # The added loader by automatically refresh, using file creation time as the query time
                    self._add_loader(loader)
                else:
                    self._loader_pool[summary_path].query_time = query_time
                    self._loader_pool.move_to_end(summary_path, last=True)

    def _generate_loader_from_relative_path(self, relative_path: str) -> ExplainLoader:
        """Generate explain loader from the given relative path."""
        self._check_summary_exist(relative_path)
        current_dir = os.path.realpath(FileHandler.join(self._summary_base_dir, relative_path))
        loader_id = self._generate_loader_id(relative_path)
        loader = ExplainLoader(loader_id=loader_id, summary_dir=current_dir)
        return loader

    def _add_loader(self, loader):
        """Add loader to the loader_pool."""
        if loader.train_id not in self._loader_pool:
            self._loader_pool[loader.train_id] = loader
        else:
            self._loader_pool.move_to_end(loader.train_id, last=True)

        while len(self._loader_pool) > self._max_loaders_num:
            self._loader_pool.popitem(last=False)

    def _execute_loading(self):
        """Execute the data loading."""
        # We will load the newest loader first.
        for loader_id in list(self._loader_pool.keys())[::-1]:
            with self._loader_pool_mutex:
                loader = self._loader_pool.get(loader_id, None)
                if loader is None:
                    logger.debug('Loader %r has been deleted, will not load data.', loader_id)
                    continue

            if self.status == _ExplainManagerStatus.STOPPING.value:
                logger.info('Loader %s status is %s, will return.', loader_id, loader.status)
                return

            loader.load()

    def _delete_loader(self, loader_id):
        """Delete loader given loader_id."""
        if loader_id in self._loader_pool:
            self._loader_pool.pop(loader_id)
            logger.debug('delete loader %s, and stop this loader loading data.', loader_id)

    def _check_status_valid(self):
        """Check manager status."""
        if self.status == _ExplainManagerStatus.INIT.value:
            raise exceptions.SummaryLogIsLoading('Data is loading, current status is %s' % self.status)

    def _check_summary_exist(self, loader_id):
        """Verify thee train_job is existed given loader_id."""
        if not self._summary_watcher.is_summary_directory(self._summary_base_dir, loader_id):
            raise ParamValueError('Can not find the train job in the manager.')

    def _reload_data_again(self):
        """Reload the data one more time."""
        logger.debug('Start to reload data again.')

        def _wrapper():
            if self.status == _ExplainManagerStatus.STOPPING.value:
                return
            self._stop_load_data()
            self._load_data()

        thread = threading.Thread(target=_wrapper, name='explainer.reload_data_thread')
        thread.daemon = False
        thread.start()

    def _stop_load_data(self):
        """Stop loading data, status changes to Stopping."""
        if self.status != _ExplainManagerStatus.LOADING.value:
            return

        logger.info('Start to stop loading data, set status to %s.', _ExplainManagerStatus.STOPPING.value)
        self.status = _ExplainManagerStatus.STOPPING.value

        for loader in self._loader_pool.values():
            loader.stop()

        while self.status != _ExplainManagerStatus.DONE.value:
            continue
        logger.info('Stop loading data end.')

    @property
    def status(self):
        """Get the status of this manager with lock."""
        with self._status_mutex:
            return self._loading_status

    @status.setter
    def status(self, status):
        """Set the status of this manager with lock."""
        with self._status_mutex:
            self._loading_status = status

    @staticmethod
    def _generate_loader_id(relative_path):
        """Generate loader id for given path"""
        loader_id = relative_path
        return loader_id


EXPLAIN_MANAGER = ExplainManager(summary_base_dir=settings.SUMMARY_BASE_DIR)



# Copyright 2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Debugger python API."""
import os.path
from typing import Iterable
import csv
import stat
from pathlib import Path
import numpy as np

import mindinsight
from mindinsight.debugger.api.conditions import WatchpointHit, HitDetail, WatchpointHandle, WatchpointHitImpl
from mindinsight.debugger.api.debugger_engine import DebuggerEngine
from mindinsight.debugger.api.debugger_tensor import DebuggerTensor, DebuggerTensorImpl
from mindinsight.debugger.api.node import Node, NodeImpl, NodeUniqueId
from mindinsight.debugger.api.statistic import TensorStatistic, SummaryStatistic
from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError
from mindinsight.debugger.common.log import LOGGER as log
from mindinsight.debugger.common.log import setup_logger
from mindinsight.debugger.common.utils import (
    validate_type, validate_slots, parse_param_to_iterable_obj
)
from mindinsight.common.util import version_match
from mindinsight.debugger.dump.parser import DebuggerParser
from mindinsight.debugger.stream_cache.data_loader import DataLoader
from mindinsight.domain.graph.base import NodeType
from mindinsight.domain.graph.query import construct_filter_func
from mindinsight.utils.exceptions import VersionNotMatchError


class DumpAnalyzer:
    """
    Analyzer to inspect the dump data.

    .. warning::
        All APIs in this class are experimental prototypes that are subject to
        change or deletion.

    Args:
        dump_dir (str): The path of the dump folder.
        mem_limit (int, optional): The memory limit for checking watchpoints in MB.
            Optional values: from 2048 MB to 2147483647 MB. ``None`` means no limit is set,
            only limited by computor memory. Default: ``None``.

    Supported Platforms:
        ``Ascend`` ``GPU``

    Examples:
            >>> from mindinsight.debugger import DumpAnalyzer
            >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
    """

    def __init__(self, dump_dir, mem_limit=None):
        self._dump_dir = os.path.realpath(dump_dir)
        self._mem_limit = 0 if mem_limit is None else mem_limit
        self._data_loader = None
        self._debuger_engine = None
        self._parser = None
        # the key is rank_id, the value is <tensor_feature, TensorImpl> map
        self._nodes = {}
        self._initialize()

    def _initialize(self):
        """Initialize."""
        self._validate_mem_limit(self._mem_limit)
        self._data_loader = DataLoader(self._dump_dir)
        self._debugger_engine = DebuggerEngine(self._data_loader, self._mem_limit)
        self._check_version()
        self._parse()

    @staticmethod
    def _validate_mem_limit(mem_limit):
        """Validate memory limit."""
        validate_type(mem_limit, 'mem_limit', int, 'int or None')
        # The unit is MB, from 2G to max value of int32 MB
        min_limit_value = 2 * 1024
        max_limit_value = 2147483647
        if mem_limit and mem_limit < min_limit_value or mem_limit > max_limit_value:
            msg = f"If mem_limit is not None, it should be set in [{min_limit_value}, {max_limit_value}]."
            raise DebuggerParamValueError(msg)

    def _check_version(self):
        """Check version."""
        dbg_services_module = self._debugger_engine.dbg_services_module
        console = setup_logger('debugger', 'console', console=True, logfile=False, formatter='%(message)s')
        ms_version = dbg_services_module.get_version()
        mi_version = mindinsight.__version__
        version_match_stat = version_match(ms_version, mi_version)
        if version_match_stat > 1:
            raise VersionNotMatchError(f"[WARNING] Current version of MindSpore({ms_version}) "
                                       f"is not compatible with MindInsight({mi_version}). "
                                       f"Otherwise some functions might not "
                                       f"work or even raise error. Please make MindSpore "
                                       f"version equal to MindInsight`s.")
        if version_match_stat == 1:
            console.warning("[WARNING] Current version of MindSpore(%s) "
                            "is not completely compatible with MindInsight(%s). "
                            "Otherwise some functions might not work.", ms_version, mi_version)

        config_json = self._data_loader.get_config_json_data()
        ms_data_version = config_json.get("ms_version", None)
        data_match_stat = version_match(ms_data_version, mi_version)
        if data_match_stat > 0:
            console.warning("[WARNING] The summary data under the `dump_dir` from MindSpore(%s) "
                            "which the version should be equal to MindInsight`s(%s) . "
                            "Otherwise some functions might not "
                            "work or even raise error.", ms_data_version, mi_version)

    def _parse(self):
        """Parse graph into nodes and tensors."""

        def _add_node_impl(base_nodes, node_type, nodes):
            nonlocal id_to_name_map
            for b_node in base_nodes:
                new_node = NodeImpl(b_node, node_type)
                new_node.debugger_engine = self._debugger_engine
                nodes[new_node.rank][new_node.unique_id] = new_node
                id_to_name_map[new_node.rank][(b_node.graph_name, b_node.name)] = new_node.name

        def _update_node_list(node_list, dst_id, cur_node, cur_node_map):
            nonlocal id_to_name_map
            dst_node_name = id_to_name_map[cur_node.rank].get(dst_id)
            if not dst_node_name:
                log.info("Failed to find %s in id_to_name_map", dst_id)
                return
            unique_id = NodeUniqueId(name=dst_node_name,
                                     rank=cur_node.rank, graph_name=cur_node.graph_name)
            target_node = cur_node_map.get(unique_id)
            if not target_node:
                log.error("Failed to find %s in node_map", unique_id)
                return
            node_list.append(target_node)

        self._parser = DebuggerParser(self._data_loader)
        ranks = self.get_ranks()
        # the key is rank_id, the value is <node_unique_id, NodeImpl> map
        self._nodes = {rank_id: {} for rank_id in ranks}
        id_to_name_map = {rank_id: {} for rank_id in ranks}
        _add_node_impl(self._parser.constants, NodeType.CONSTANT, self._nodes)
        _add_node_impl(self._parser.parameters, NodeType.PARAMETER, self._nodes)
        _add_node_impl(self._parser.operators, NodeType.OPERATOR, self._nodes)
        # update input and output nodes
        for node_map in self._nodes.values():
            for node in node_map.values():
                base_node = node.base_node
                if hasattr(base_node, 'inputs'):
                    # parameter or const node has no inputs
                    for node_input in base_node.inputs:
                        _update_node_list(node.input_nodes, (base_node.graph_name, node_input.name), node, node_map)

        for node_map in self._nodes.values():
            for node in node_map.values():
                for node_input in node.input_nodes:
                    node_input.downstream.append(node)

    def export_graphs(self, output_dir=None):
        """
        Export the computational graph(s) in xlsx file(s) to the `output_dir` .

        The file(s) will contain the stack info of graph nodes.

        Args:
            output_dir (str, optional): Output directory to save the file.
                ``None`` means to use the current working directory. Default: ``None``.

        Returns:
            str, The path of the generated file.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> res = my_run.export_graphs()
        """
        return self._parser.export_xlsx(output_dir)

    def select_nodes(
            self,
            query_string,
            use_regex=False,
            select_by="node_name",
            ranks=None,
            case_sensitive=True) -> Iterable[Node]:
        """
        Select nodes.

        Select the matched nodes in the computational graph according to the
        query_string. The nodes can be matched by "node_name" or "code_stack",
        see the parameters for detail.

        Args:
            query_string (str): Query string. For a node to be selected, the
                match target field must contains or matches the query string.
            use_regex (bool, optional): Indicates whether query is a regex. Default:
                ``False``.
            select_by (str, optional): The field to search when selecting
                nodes. Available values are ``"node_name"``, ``"code_stack"``.
                ``"node_name"`` means to search the name of the nodes in the
                graph. ``"code_stack"`` means the stack info of
                the node. Default: ``"node_name"``.
            ranks (Union[int, list[int], None], optional): The rank(s) to select. ``None`` means all ranks will be
                considered. The selected nodes must exist on the specified ranks. Default: ``None``.
            case_sensitive (bool, optional): Whether case-sensitive when
                selecting tensors. Default: ``True``.

        Returns:
            Iterable[Node], the matched nodes.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> nodes = my_run.select_nodes("Conv2D-op13")
        """
        validate_type(query_string, 'query_string', str, 'str')
        validate_type(use_regex, 'use_regex', bool, 'bool')
        validate_type(case_sensitive, 'case_sensitive', bool, 'bool')
        node_filter_func = self._get_filter_func(select_by)
        ranks = self._get_iterable_ranks(ranks)

        query_filter_func = construct_filter_func(query_string, case_sensitive, use_regex)
        nodes = []
        for rank_id in ranks:
            for node in self._nodes.get(rank_id, {}).values():
                if node_filter_func(node, query_filter_func):
                    nodes.append(node.copy())
        return nodes

    @staticmethod
    def _match_name(node, filter_func):
        """Check if name matched."""
        return filter_func(node.name)

    @staticmethod
    def _match_stack(node, filter_func):
        """Check if stack matched."""
        for res in map(filter_func, node.stack):
            if res:
                return True
        return False

    @staticmethod
    def _get_filter_func(select_by):
        """Get filter function."""
        if select_by == 'node_name':
            return DumpAnalyzer._match_name
        if select_by == 'code_stack':
            return DumpAnalyzer._match_stack
        raise DebuggerParamValueError(
            "The param `select_by` only support `node_name` or `code_stack`.")

    def select_tensors(
            self,
            query_string,
            use_regex=False,
            select_by="node_name",
            iterations=None,
            ranks=None,
            slots=None,
            case_sensitive=True) -> Iterable[DebuggerTensor]:
        """
        Select tensors.

        Select the matched tensors in the directory according to the
        sepicified filter condition, see the parameters for detail.

        Args:
            query_string (str): Query string. For a tensor to be selected, the
                match target field must contain or match the query string.
            use_regex (bool, optional): Indicates whether query is a regex. Default:
                ``False``.
            select_by (str, optional): The field to search when selecting
                tensors. Available values are ``"node_name"``, ``"code_stack"``.
                ``"node_name"`` means to search the node name of the tensors in the
                graph. ``"code_stack"`` means the stack info of
                the node that outputs this tensor. Default: ``"node_name"``.
            iterations (Union[int, list[int], None], optional): The iteration(s) to select. ``None`` means all dumped
                iterations will be selected. Default: ``None``.
            ranks (Union[int, list[int], None], optional): The rank(s) to select. ``None`` means all ranks
                will be selected. Default: ``None``.
            slots (list[int], optional): The slot of the selected tensor. ``None`` means all slots will be selected.
                Default: ``None``.
            case_sensitive (bool, optional): Whether case-sensitive when selecting tensors. Default: ``True``.

        Returns:
          Iterable[DebuggerTensor], the matched tensors.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> tensors = my_run.select_tensors("Conv2D-op13")
        """
        validate_type(query_string, 'query_string', str, 'str')
        validate_type(use_regex, 'use_regex', bool, 'bool')
        validate_type(case_sensitive, 'case_sensitive', bool, 'bool')
        validate_slots(slots)
        node_filter_func = self._get_filter_func(select_by)
        ranks = self._get_iterable_ranks(ranks)
        dumped_iterations = self.get_iterations(ranks)
        iterations = parse_param_to_iterable_obj(iterations, 'iterations', dumped_iterations)

        tensors = []
        query_filter_func = construct_filter_func(query_string, case_sensitive, use_regex)
        for rank_id in ranks:
            for node in self._nodes.get(rank_id, {}).values():
                if node_filter_func(node, query_filter_func):
                    tensors.extend(node.get_output_tensors(slots=slots, iterations=iterations))
        return tensors

    def select_tensor_statistics(
            self,
            iterations=None,
            ranks=None):
        """
        Select tensor statistics.

        Select the matched tensor statistics in the directory according to the
        sepicified filter condition, see the parameters for detail.

        Args:
            iterations (Union[int, list[int], None], optional): The iteration(s) to select. ``None`` means all dumped
                iterations will be selected. Default: ``None``.
            ranks (Union[int, list[int], None], optional): The rank(s) to select. ``None`` means all ranks
                will be selected. Default: ``None``.

        Returns:
            Dict[TensorStatistic], the matched TensorStatistics. The format is as below.

            .. code-block::

                {
                "rank_id":
                    {
                    "iteration_id":
                        {
                        "tensor_name":
                            [TensorStatistic],
                        ...
                        }
                    }
                ...
                }

        Examples:
            >>> from mindinsight.debugger import DumpAnalyzer
            >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
            >>> statistics = my_run.select_tensor_statistics(ranks=[0])
        """
        ranks = self._get_iterable_ranks(ranks)
        dumped_iterations = self.get_iterations(ranks)
        iterations = parse_param_to_iterable_obj(iterations, 'iterations', dumped_iterations)

        tensor_statistics = {}
        for rank_id in ranks:
            static_in_rank = {}
            for iteration_id in self.get_iterations(ranks=[rank_id]):
                if iteration_id not in iterations:
                    continue
                tensor_statistic = self._get_statistic(rank_id=rank_id, iteration_id=iteration_id)
                static_in_rank[iteration_id] = tensor_statistic
            tensor_statistics[rank_id] = static_in_rank
        return tensor_statistics

    def _get_statistic(self, rank_id, iteration_id):
        """
        Get the TensorStatic of the corresponding rank and iteration.
        """
        tensor_statistics = {}
        if not self._data_loader.has_data:
            return tensor_statistics
        net_name = self._data_loader.get_net_name()
        net_dir = Path(os.path.join(self._dump_dir, 'rank_' + str(rank_id), net_name)).absolute()
        for graph_dir in net_dir.iterdir():
            target_step_dir = graph_dir / str(iteration_id)
            if not target_step_dir.is_dir():
                continue
            statistic_file_path = os.path.join(target_step_dir, "statistic.csv")
            with open(statistic_file_path, 'r') as f:
                csv_reader = csv.DictReader(f)
                # The first line of the csv file is the title, so skip the first line.
                for statistic_line in csv_reader:
                    tensor_name = statistic_line.get('Op Name') + ':' + statistic_line.get(
                        'IO') + ':' + statistic_line.get('Slot')
                    statistics_for_specified_name = tensor_statistics.get(tensor_name, [])
                    statistics_for_specified_name.append(statistic_line)
                    tensor_statistics.update({tensor_name: statistics_for_specified_name})
        return tensor_statistics

    def _compute_statistics(self, debugger_tensors):
        """
        Compute the statistic of the given tensors.

        Args:
            debugger_tensors(Iterable[DebuggerTensor]): The given DebuggerTensors.

        Returns:
            Dict[TensorStatistic], the computed TensorStatistics. The format is as below.

            .. code-block::

                {"rank_id":{
                    "iteration_id":
                        {
                        "tensor_name":
                            [TensorStatistic],
                        ...
                        }
                    }
                ...
                }
        """
        statistics = {}
        for tensor in debugger_tensors:
            rank_id = tensor.rank
            is_new_rank = rank_id not in statistics
            static_in_rank = statistics.get(rank_id, {})
            iteration = tensor.iteration
            is_new_iteration = iteration not in static_in_rank
            static_in_iter = static_in_rank.get(iteration, {})
            single_static = self._compute_statistic(tensor)
            tensor_name = single_static.op_name + ':' + single_static.io + ':' + str(single_static.slot)
            statistics_for_specified_name = static_in_iter.get(tensor_name, [])
            statistics_for_specified_name.append(single_static)
            static_in_iter.update({tensor_name: statistics_for_specified_name})
            if is_new_iteration:
                static_in_rank[iteration] = static_in_iter
            if is_new_rank:
                statistics[rank_id] = static_in_rank
        return statistics

    def _compute_statistic(self, debugger_tensor):
        """Compute the tensor statistic for one tensor."""
        statistic = TensorStatistic()
        op_name = debugger_tensor.node.name
        tensor_value = debugger_tensor.value()
        if tensor_value is None:
            return statistic
        short_name = op_name.split('/')[-1]
        op_type = short_name.split('-')[0]
        statistic.op_name = short_name
        statistic.op_type = op_type
        statistic.io = 'output'
        statistic.slot = debugger_tensor.slot
        statistic.data_size = tensor_value.nbytes
        statistic.data_type = tensor_value.dtype
        statistic.shape = tensor_value.shape
        statistic.min_value = np.amin(tensor_value)
        statistic.max_value = np.amax(tensor_value)
        statistic.avg_value = tensor_value.mean()
        statistic.count = tensor_value.size
        statistic.negative_zero_count = np.sum(np.where(tensor_value == 0, 1, 0))
        statistic.positive_zero_count = np.sum(np.where(tensor_value == 0, 1, 0))
        statistic.negative_inf_count = len(np.nonzero(np.isneginf(tensor_value))[0])
        statistic.positive_inf_count = len(np.nonzero(np.isposinf(tensor_value))[0])
        statistic.nan_count = len(np.nonzero(np.isnan(tensor_value))[0])
        return statistic

    def summary_statistics(self, statistics, overflow_value=65500, out_path="./"):
        """
        Summary the statistics in the different ranks and iterations.

        Args:
            statistics(Dict[TensorStatistic]): The given TensorStatistic. They can be the return value of
                `compute_statistic` or `select_tensor_statistics`.
            overflow_value(int, optional): The given overflow threshold, default: ``65500``.
            out_path(str, optional): The given output directory to save the statistics. Default: ``"./"``.
        """
        summary_statistics = {}
        for rank_id, statistics_in_rank in statistics.items():
            log.warning("process statistics in rank, rank_id is: %s", rank_id)
            for iteration_id, statistics_in_iteration in statistics_in_rank.items():
                log.warning("process statistics in iteration, iteration_id is: %s", iteration_id)
                for tensor_name, statistics_for_name in statistics_in_iteration.items():
                    for statistic in statistics_for_name:
                        if isinstance(statistic, TensorStatistic):
                            self._put_tensor_statistic_to_summarystatistics(statistic, summary_statistics,
                                                                            overflow_value)
                        else:
                            self._put_dict_statistic_to_summarystatistics(statistic, summary_statistics, overflow_value)
                    summary_statistic = summary_statistics.get(tensor_name, SummaryStatistic())
                    summary_statistic.total_iterations += 1
        self._export_to_disk(summary_statistics, out_path)

    def _put_dict_statistic_to_summarystatistics(self, statistic, summary_statistics, overflow_value=65500):
        """Put dict_statistic to summarized statistics, used for Statistic of dict type from statistic file."""
        op_name = statistic.get('Op Name', 'unkown')
        op_type = statistic.get('Op Type', 'unkown')
        io = statistic.get('IO', 'output')
        slot = statistic.get('Slot', 0)
        tensor_name = op_name + ":" + io + ":" + str(slot)
        positive_inf_count = statistic.get('Positive Inf Count', 0)
        negative_inf_count = statistic.get('Negative Inf Count', 0)
        nan_count = statistic.get('NaN Count', 0)
        min_value = statistic.get('Min Value', 0)
        max_value = statistic.get('Max Value', 0)
        has_out_of_range = False
        summary_statistic = summary_statistics.get(tensor_name, SummaryStatistic())
        if not summary_statistic.op_type:
            summary_statistic.op_type = op_type
        if not summary_statistic.op_name:
            summary_statistic.op_name = op_name
        if not summary_statistic.tensor_name:
            summary_statistic.tensor_name = tensor_name
            summary_statistics[tensor_name] = summary_statistic
        summary_statistic.total_number += 1
        if int(positive_inf_count) > 0 or int(negative_inf_count) > 0:
            summary_statistic.inf_number += 1
        if int(nan_count) > 0:
            summary_statistic.nan_number += 1
        if abs(float(min_value)) > overflow_value or abs(float(max_value)) > overflow_value:
            summary_statistic.out_of_range_number += 1
            has_out_of_range = True
        if int(positive_inf_count) == 0 and int(negative_inf_count) == 0 and int(
                nan_count) == 0 and not has_out_of_range:
            summary_statistic.none_overflow_number += 1

    def _put_tensor_statistic_to_summarystatistics(self, statistic, summary_statistics, overflow_value=65500):
        """Put tensor_statistic to summarized statistics, used for TensorStatistic from tensor file."""
        op_name = statistic.op_name
        op_type = statistic.op_type
        io = statistic.io
        slot = statistic.slot
        min_value = statistic.min_value
        max_value = statistic.max_value
        has_out_of_range = False
        tensor_name = op_name + ":" + io + ":" + str(slot)
        summary_statistic = summary_statistics.get(tensor_name, SummaryStatistic())
        if not summary_statistic.op_type:
            summary_statistic.op_type = op_type
        if not summary_statistic.op_name:
            summary_statistic.op_name = op_name
        if not summary_statistic.tensor_name:
            summary_statistic.tensor_name = tensor_name
            summary_statistics[tensor_name] = summary_statistic
        summary_statistic.total_number += 1
        if statistic.positive_inf_count > 0 or statistic.negative_inf_count > 0:
            summary_statistic.inf_number += 1
        if statistic.nan_count > 0:
            summary_statistic.nan_number += 1
        if abs(min_value) > overflow_value or abs(max_value) > overflow_value:
            summary_statistic.out_of_range_number += 1
            has_out_of_range = True
        if statistic.positive_inf_count == 0 and statistic.negative_inf_count == 0 and statistic.nan_count == 0 and \
                not has_out_of_range:
            summary_statistic.none_overflow_number += 1

    def _export_to_disk(self, tensor_statistics, out_path="./"):
        """
        Export the tensor staticstics to the out_path.

        Args:
            tensor_statistics(Union[Dict[TensorStatistic], Dict[SummaryStatistic]]): The given Statistics.
                They can be the return value of `compute_statistic` or `summary_statistics`.
            out_path(str, optional): The given output directory to save the statistics. Default: ``"./"``.

        """
        ks = tensor_statistics.keys()
        if not ks:
            log.warning("The given tensor_statistics is empty.")
            return
        if not os.path.exists(out_path):
            os.makedirs(out_path, exist_ok=True)
        if isinstance(list(ks)[0], int):
            for rank_id, statistics_in_rank in tensor_statistics.items():
                for iteration_id, statistics_in_iteration in statistics_in_rank.items():
                    self._export_statistics_in_one_iteration(statistics_in_iteration, rank_id, iteration_id, out_path)
        elif isinstance(list(ks)[0], str):
            self._export_summary_statistics(tensor_statistics, out_path)
        else:
            log.warning("Invalid tensor_statistics data structure.")

    def _export_summary_statistics(self, tensor_statistics, out_path):
        """Export the summarized statistics to out_path."""
        statistic_file_path = os.path.join(out_path, "statistics_summary.csv")
        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
        modes = stat.S_IWUSR | stat.S_IRUSR
        with os.fdopen(os.open(statistic_file_path, flags, modes), 'w', newline='') as f:
            csv_writer = csv.writer(f)
            statistic_header = ["op_type", "op_name", "tensor_name", "inf_number", "nan_number",
                                "out_of_range_number", "total_number", "none_overflow_number", "total_iterations"]
            csv_writer.writerow(statistic_header)
            for _, statistic in tensor_statistics.items():
                statistic_line = [statistic.op_type, statistic.op_name, statistic.tensor_name,
                                  statistic.inf_number, statistic.nan_number, statistic.out_of_range_number,
                                  statistic.total_number, statistic.none_overflow_number, statistic.total_iterations]
                csv_writer.writerow(statistic_line)
        log.info("export summarised statistics to file: %s", statistic_file_path)

    def _export_statistics_in_one_iteration(self, tensor_statistics_in_one_iteration, rank_id, iteration_id, out_path):
        """Export tensor_statistics in one iteration."""
        iteration_path = os.path.join(out_path, str(rank_id), str(iteration_id))
        if not os.path.exists(iteration_path):
            os.makedirs(iteration_path, exist_ok=True)
        statistic_file_path = os.path.join(iteration_path, "statistics.csv")
        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
        modes = stat.S_IWUSR | stat.S_IRUSR
        with os.fdopen(os.open(statistic_file_path, flags, modes), 'w', newline='') as f:
            csv_writer = csv.writer(f)
            statistic_header = ["Op Type", "Op Name", "Task ID", "Stream ID", "Timestamp", "IO", "Slot", "Data Size",
                                "Data Type", "Shape", "Max Value", "Min Value", "Avg Value", "Count",
                                "Negative Zero Count", "Positive Zero Count", "NaN Count", "Negative Inf Count",
                                "Positive Inf Count", "Zero Count"]
            csv_writer.writerow(statistic_header)
            for statistic in tensor_statistics_in_one_iteration:
                statistic_line = [statistic.op_type, statistic.op_name, statistic.task_id,
                                  statistic.stream_id, statistic.time_stamp, statistic.io, statistic.slot,
                                  statistic.data_size, statistic.data_type, statistic.shape, statistic.max_value,
                                  statistic.min_value, statistic.avg_value, statistic.count,
                                  statistic.negative_zero_count, statistic.positive_zero_count, statistic.nan_count,
                                  statistic.negative_inf_count, statistic.positive_inf_count, statistic.zero_count]
                csv_writer.writerow(statistic_line)
        log.info("export summarised statistics to file: %s", statistic_file_path)

    def get_iterations(self, ranks=None) -> Iterable[int]:
        """
        Get available iterations which have data dumped in this run.

        Args:
            ranks (Union[int, list[int], None], optional): The rank(s) to select.
                Get available iterations which are under the specified ranks.
                The ranks refers to the number of devices to be used starting from 0
                when running distributed training. This number is called rank.
                For example, for an 8-card computer, only 4-7 cards are used for
                specified training, so 4-7 cards correspond to the ranks 0-3 respectively..
                If ``None``, return iterations of all ranks. Default: ``None``.

        Returns:
            Iterable[int], available iterations which have dumped data, sorted in increasing order.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> iterations = my_run.get_iterations()
                >>> print(list(iterations))
                [0]
        """
        total_dumped_steps = self._data_loader.load_dumped_step()
        ranks = self._get_iterable_ranks(ranks)
        iterations = set()
        for rank_id in ranks:
            for iters_per_graph in total_dumped_steps.get(rank_id, {}).values():
                iterations.update(iters_per_graph)
        res = list(iterations)
        res.sort()
        return res

    def get_ranks(self) -> Iterable[int]:
        """
        Get the available ranks in this run.

        Returns:
            Iterable[int], the list of rank id in current dump directory.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> ranks = my_run.get_ranks()
                >>> print(list(ranks))
                [0]
        """
        return [rank_dir.rank_id for rank_dir in self._data_loader.rank_dirs]

    def check_watchpoints(
            self,
            watchpoints,
            error_on_no_value=False) -> Iterable[WatchpointHit]:
        """
        Check the given watchpoints in a batch.

        Note:
            1. For speed, all watchpoints for the iteration should be given at
            the same time to avoid reading tensors len(watchpoints) times.

            2. The `check_watchpoints` function start a new process when it is called, needs to be
            called in `if __name__ == '__main__'` .

        Args:
            watchpoints (Iterable[Watchpoint]): The list of watchpoints.
            error_on_no_value (bool, optional): Whether to report error code in watchpoint
                hit when the specified tensor have no value stored in
                dump_dir. Default: ``False``.

        Returns:
            Iterable[WatchpointHit], the watchpoint hit list, sorted by tensor drop time.


        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> from mindinsight.debugger import (TensorTooLargeCondition,
                ...                                    Watchpoint)
                >>>
                >>> def test_watchpoints():
                ...     my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                ...     tensors = my_run.select_tensors(
                ...                                         query_string="Conv2D-op13",
                ...                                         use_regex=True,
                ...                                         iterations=[0],
                ...                                         ranks=[0],
                ...                                         slots=[0]
                ...                                         )
                ...     watchpoint = Watchpoint(tensors=tensors,
                ...                             condition=TensorTooLargeCondition(abs_mean_gt=0.0))
                ...     # the check_watchpoints function start a new process needs to be called through the main entry
                ...     hit = list(my_run.check_watchpoints(watchpoints=[watchpoint]))[0]
                ...     # print(str(hit))
                ...     # the print result is as follows
                ...     # Watchpoint TensorTooLarge triggered on tensor:
                ...     # rank: 0
                ...     # graph_name: kernel_graph_0
                ...     # node_name: Default/network-WithLossCell/_backbone-AlexNet/conv2-Conv2d/Conv2D-op13
                ...     # slot: 0
                ...     # iteration: 0
                ...     # Threshold: {'abs_mean_gt': 0.0}
                ...     # Hit detail: the setting for watchpoint is abs_mean_gt = 0.0.
                ...     # The actual value of the tensor is abs_mean_gt = 0.06592023578438996.
                ...
                >>> if __name__ == "__main__":
                ...     test_watchpoints()
                ...
        """
        wp_hit_list = []
        # key is watchpoint_id, value is a dict with iteration as the key and check_nodes as values
        iterations = set()
        wp_handles = {wp_id: WatchpointHandle(wp_id, wp) for wp_id, wp in enumerate(watchpoints)}
        for wp_handle in wp_handles.values():
            iterations.update(wp_handle.get_iterations())
        debugger_backend = self._debugger_engine.dbg_service
        # check all the watchpoint for the iterations
        for iteration in iterations:
            log.info("Check watchpoints for iteration %s", iteration)
            if not self._data_loader.has_data(iteration):
                log.info("No data dumped with iteration id: %s. Ignore checking watchpoint.", iteration)
                continue
            # adding the watchpoint for current iteration
            for wp_handle in wp_handles.values():
                wp_handle.add_watchpoint(iteration, self._debugger_engine)
            # check the watchpoint for current iteration
            # add the hit watchpoints to hit list
            hit_list = debugger_backend.check_watchpoints(
                iteration=iteration, error_on_no_value=error_on_no_value)
            for hit in hit_list:
                # the list of slots for the hit node to report
                # (for the current watchpoint and iteration)
                wp_handle = wp_handles.get(hit.watchpoint_id)
                graph_name, node_name = hit.name.split('/', 1)
                node = self._get_node(node_name, hit.rank_id, graph_name)
                tensor = DebuggerTensorImpl(node=node, slot=hit.slot, iteration=iteration)
                if wp_handle.need_check(tensor):
                    hit_params = hit.parameters
                    hit_detail = HitDetail(hit_params, wp_handle.condition)
                    wp_hit = WatchpointHitImpl(tensor=tensor,
                                               condition=wp_handle.condition,
                                               hit_detail=hit_detail,
                                               error_code=hit.error_code)
                    wp_hit_list.append(wp_hit)
            if error_on_no_value:
                no_value_hit_list = []
                for wp_handle in wp_handles.values():
                    no_value_hit_list += wp_handle.watchpoint_hit_on_no_value(iteration)
                wp_hit_list += no_value_hit_list
            # remove all the watchpoints for the previous iterations
            for watchpoint_id in wp_handles:
                debugger_backend.remove_watchpoint(watchpoint_id=watchpoint_id)

        return wp_hit_list

    def _get_node(self, node_name, rank_id, graph_name):
        """Get NodeImpl object."""
        unique_id = NodeUniqueId(name=node_name, rank=rank_id, graph_name=graph_name)
        node = self._nodes.get(rank_id, {}).get(unique_id)
        return node

    def list_affected_nodes(self, tensor):
        """
        List the nodes that use given tensor as input.

        Affected nodes is defined as the nodes use the given tensor as input. If
        a node is affected by the given tensor, the node's output value is
        likely to change when the given tensor changes.

        Args:
            tensor (DebuggerTensor): The tensor of which affected nodes will be
                returned.

        Returns:
            Iterable[Node], the affected nodes of the specified tensor.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> tensor_list = list(my_run.select_tensors(query_string="Conv2D-op13"))
                >>> affected_nodes = my_run.list_affected_nodes(tensor_list[0])
        """
        self._validate_node(tensor.node)
        affected_nodes = [affected_node.copy() for affected_node in tensor.node.downstream]
        return affected_nodes

    def get_input_nodes(self, node):
        """
        Get the input nodes of the given node.

        Args:
            node (Node): The node of which input nodes will be returned.

        Returns:
            Iterable[Node], the input nodes of the specified node.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> node_list = list(my_run.select_nodes(query_string="Conv2D-op13"))
                >>> input_nodes = my_run.get_input_nodes(node_list[0])
        """
        self._validate_node(node)
        input_nodes = node.input_nodes.copy()
        return input_nodes

    def get_output_nodes(self, node):
        """
        Get the nodes that use the output tensors of the given node.

        Args:
            node (Node): The specified node.

        Returns:
            Iterable[Node], output nodes of the specified node.

        Examples:
                >>> from mindinsight.debugger import DumpAnalyzer
                >>> my_run = DumpAnalyzer(dump_dir="/path/to/your/dump_dir_with_dump_data")
                >>> node_list = list(my_run.select_nodes(query_string="Conv2D-op13"))
                >>> out_nodes = my_run.get_output_nodes(node_list[0])
        """
        self._validate_node(node)
        output_nodes = node.downstream.copy()
        return output_nodes

    def _validate_node(self, node):
        """Check if node is in current directory."""
        validate_type(node, 'node', NodeImpl, 'Node')
        node = self._get_node(node.name, node.rank, node.graph_name)
        if node is None:
            raise DebuggerParamValueError(f"Failed to find node {node.name} with rank {node.rank} "
                                          "in dump directory.")

    def _get_iterable_ranks(self, ranks):
        """
        Validate input ranks and return iterable rands.

        Args:
           ranks (Union[int, list[int], None], optional): The range of ranks.

        Returns:
            list[int], list of rank id.
        """
        total_ranks = self.get_ranks()
        return parse_param_to_iterable_obj(ranks, 'ranks', total_ranks)



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The common proposer."""
from collections import OrderedDict

from mindinsight.profiler.common.util import get_options
from mindinsight.profiler.proposer.allproposers.base_proposer import Proposer
from mindinsight.profiler.common.log import logger


class CommonProposer(Proposer):
    """The common proposer."""
    def __init__(self, profiling_dir, rank_id):
        super().__init__(profiling_dir, rank_id)
        self.__proposal_dict = OrderedDict()
        self.__proposal_dict["common-profiler_tutorial"] = None

    def analyze(self, options=None):
        """
        Get the proposal from proposer.

        Args:
            options: options for proposer analysis

        Returns:
            dict, the proposal from proposer instance.
        """
        logger.info('The CommonProposer　is running')
        options = get_options(options)
        return self.__proposal_dict



# Copyright 2019-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Config file for gunicorn."""

import os
import time
import signal
import sys
import multiprocessing
import threading
from importlib import import_module

import psutil
import gunicorn

from mindinsight.utils.computing_resource_mgr import terminate
from mindinsight.debugger.session_manager import SessionManager


gunicorn.SERVER_SOFTWARE = 'unknown'
worker_class = 'sync'
# number of parallel worker processes
workers = 1
threads = min(30, os.cpu_count() * 2 + 1)
worker_connections = 1000

timeout = 30
graceful_timeout = 30
daemon = False

captureoutput = False

# write gunicorn default log to devnull, and using mindinsight logger write gunicorn log to file.
accesslog = os.devnull


def on_starting(server):
    """Hook function on starting gunicorn process."""
    hook_module = import_module('mindinsight.utils.hook')
    for hook in hook_module.HookUtils.instance().hooks():
        threading.Thread(target=hook.on_startup, args=(server.log,)).start()


# This global variable is to manage the listen process so that we can close the
# process when gunicorn is exiting.
LISTEN_PROCESS = None


def post_worker_init(worker):
    """
    Launch a process to listen worker after gunicorn worker is initialized.

    Children processes of gunicorn worker should be killed when worker has been killed
    because gunicorn master murders this worker for some reasons such as worker timeout.

    Args:
        worker (ThreadWorker): worker instance.
    """
    def murder_worker_children_processes():
        signal.signal(
            signal.SIGTERM,
            lambda signal_num, handler: sys.exit(0))
        processes_to_kill = []
        # sleep 3 seconds so that all worker children processes have been launched.
        time.sleep(3)
        process = psutil.Process(worker.pid)
        for child in process.children(recursive=True):
            if child.pid != os.getpid():
                processes_to_kill.append(child)
        while True:
            if os.getppid() != worker.pid:
                # Kill the remaining sub-processed after the worker process died
                _, alive = psutil.wait_procs(processes_to_kill, 0.1)
                current_worker_pid = os.getppid()
                for proc in alive:
                    worker.log.info("Original worker pid: %d, current worker pid: %d, stop process %d",
                                    worker.pid, current_worker_pid, proc.pid)
                    try:
                        proc.send_signal(signal.SIGKILL)
                    except psutil.NoSuchProcess:
                        continue
                    except psutil.Error as ex:
                        worker.log.error("Stop process %d failed. Detail: %s.", proc.pid, str(ex))
                worker.log.info("%d processes have been terminated by listener.", len(alive))
                break
            time.sleep(1)

    listen_process = multiprocessing.Process(target=murder_worker_children_processes,
                                             name="murder_worker_children_processes")
    listen_process.start()
    global LISTEN_PROCESS
    LISTEN_PROCESS = listen_process
    worker.log.info("Service start state: success.")
    worker.log.info("Server pid: %d, start to listening.", worker.ppid)


def worker_int(worker):
    """Terminate child processes when worker is interrupted."""
    terminate()
    global LISTEN_PROCESS
    if LISTEN_PROCESS is not None:
        LISTEN_PROCESS.terminate()
    SessionManager.get_instance().exit()
    worker.log.info("Worker int processed.")



# Copyright 2019-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Constants module for mindinsight settings."""
import logging
import multiprocessing


####################################
# Global default settings.
####################################
LOG_FORMAT = '[%(levelname)s] MI(%(process)d:%(thread)d,%(processName)s):%(asctime)s ' \
             '[%(filepath)s:%(lineno)d][%(sub_module)s] %(message)s'

GUNICORN_ACCESS_FORMAT = "'%(t)s %(h)s <%(r)s> %(s)s %(b)s <%(f)s> <%(a)s> %(L)s '"

LOG_LEVEL = logging.INFO
# rotating max bytes, default is 50M
LOG_ROTATING_MAXBYTES = 52428800

# rotating backup count, default is 30
LOG_ROTATING_BACKUPCOUNT = 30

####################################
# Web default settings.
####################################
HOST = '127.0.0.1'

# Allow to support cross origin resource sharing(CORS) enable. Default is disable.
# You can follow these steps to enable CORS:
# 1. Set ENABLE_CORS to True
# 2. Install `Flask-Cors` by pip, `pip install Flask-Cors>=3.0.8`
ENABLE_CORS = False

SUPPORT_REQUEST_METHODS = {'POST', 'GET', 'PUT', 'DELETE'}

# api prefix should not end with slash, correct format is /v1/url
API_PREFIX = '/v1/mindinsight'

####################################
# Datavisual default settings.
####################################
MAX_PROCESSES_COUNT = max(min(int(multiprocessing.cpu_count() * 0.75), 45), 1)

MAX_TAG_SIZE_PER_EVENTS_DATA = 300
DEFAULT_STEP_SIZES_PER_TAG = 500

MAX_GRAPH_TAG_SIZE = 10
MAX_TENSOR_TAG_SIZE = 6
MAX_IMAGE_STEP_SIZE_PER_TAG = 10
MAX_RELOAD_INTERVAL = 300
MAX_SCALAR_STEP_SIZE_PER_TAG = 1000
MAX_GRAPH_STEP_SIZE_PER_TAG = 1
MAX_HISTOGRAM_STEP_SIZE_PER_TAG = 50
MAX_TENSOR_STEP_SIZE_PER_TAG = 20
MAX_TENSOR_RESPONSE_DATA_SIZE = 100000

ENABLE_RECOMMENDED_WATCHPOINTS = True



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Condition list.

This module provide the detail conditions list.
"""
from mindinsight.debugger.conditionmgr.condition import Condition
from mindinsight.debugger.conditionmgr.condition import OptimizePhaseEnum
from mindinsight.debugger.conditionmgr.condition import ConditionParameter
from mindinsight.debugger.conditionmgr.condition import ValueTypeEnum
from mindinsight.debugger.conditionmgr.condition import TargetTypeEnum
from mindinsight.debugger.conditionmgr.condition import PlatformEnum
from mindinsight.debugger.conditionmgr.condition import ParamTypeEnum
from mindinsight.debugger.conditionmgr.condition import ConditionIdEnum
from mindinsight.debugger.conditionmgr.condition import ParamNameEnum
from mindinsight.debugger.conditionmgr.condition import check_initialization_available
from mindinsight.debugger.conditionmgr.condition import check_normal_param_range
from mindinsight.debugger.conditionmgr.condition import check_percentage_param_range
from mindinsight.debugger.conditionmgr.condition import check_abs_param_range, check_positive_param_range


CONDITION_LIST = [
    Condition(
        condition_id=ConditionIdEnum.WEIGHT_INITIALIZATION,
        abbr="WI",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_initialization
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ZERO_PERCENTAGE_GE,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_percentage_param_range,
                default_value=100
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MIN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            )
        ],
        supported_target_type=TargetTypeEnum.WEIGHT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1),
        availability_test_func=check_initialization_available
    ),
    Condition(
        condition_id=ConditionIdEnum.WEIGHT_OVERFLOW,
        abbr="WO",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_general_overflow
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[],
        supported_target_type=TargetTypeEnum.WEIGHT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.WEIGHT_TOO_LARGE,
        abbr="WL",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_too_large
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ABS_MEAN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MIN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MEAN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            )
        ],
        supported_target_type=TargetTypeEnum.WEIGHT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.WEIGHT_TOO_SMALL,
        abbr="WS",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_too_small
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ABS_MEAN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MIN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MEAN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            )
        ],
        supported_target_type=TargetTypeEnum.WEIGHT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.GRADIENT_VANISHING,
        abbr="GV",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_too_small
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ABS_MEAN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MIN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MEAN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            )
        ],
        supported_target_type=TargetTypeEnum.GRADIENT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.GRADIENT_TOO_LARGE,
        abbr="GL",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_too_large
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ABS_MEAN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MIN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MEAN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            )
        ],
        supported_target_type=TargetTypeEnum.GRADIENT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.GRADIENT_EXPLODING,
        abbr="GE",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_general_overflow
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[],
        supported_target_type=TargetTypeEnum.GRADIENT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.TENSOR_OVERFLOW,
        abbr="TO",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_general_overflow
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[],
        supported_target_type=TargetTypeEnum.TENSOR,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.OPERATOR_OVERFLOW,
        abbr="OO",
        # Send this condition to MindSpore will use WatchCondition.Condition.overflow
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[],
        supported_target_type=TargetTypeEnum.TENSOR,
        supported_platforms=(PlatformEnum.ASCEND,),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.TENSOR_TOO_LARGE,
        abbr="TL",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_too_large
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ABS_MEAN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MIN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MEAN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            )
        ],
        supported_target_type=TargetTypeEnum.TENSOR,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.TENSOR_TOO_SMALL,
        abbr="TS",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_too_small
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ABS_MEAN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MIN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MEAN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range
            )
        ],
        supported_target_type=TargetTypeEnum.TENSOR,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.TENSOR_ALL_ZERO,
        abbr="TZ",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_all_zero
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ZERO_PERCENTAGE_GE,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_percentage_param_range,
                default_value=100
            )
        ],
        supported_target_type=TargetTypeEnum.TENSOR,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.WEIGHT_NOT_CHANGED,
        abbr="WNC",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_not_changed
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.RTOL,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range,
                default_value=1e-5
            ),
            ConditionParameter(
                name=ParamNameEnum.ATOL,
                value_type=ValueTypeEnum.FLOAT64,
                support_disable=False,
                default_value=1e-8,
                visible_on_ui=False
            ),
            ConditionParameter(
                name=ParamNameEnum.EQUAL_NAN,
                value_type=ValueTypeEnum.BOOL,
                support_disable=False,
                default_value=False,
                visible_on_ui=False
            )
        ],
        supported_target_type=TargetTypeEnum.WEIGHT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.WEIGHT_CHANGE_TOO_LARGE,
        abbr="WCL",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_change_too_large
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ABS_MEAN_UPDATE_RATIO_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range,
                default_value=1e-1
            ),
            ConditionParameter(
                name=ParamNameEnum.EPSILON,
                value_type=ValueTypeEnum.FLOAT64,
                support_disable=False,
                default_value=1e-9,
                visible_on_ui=False
            )
        ],
        supported_target_type=TargetTypeEnum.WEIGHT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.WEIGHT_CHANGE_TOO_SMALL,
        abbr="WCS",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_change_too_small
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.ABS_MEAN_UPDATE_RATIO_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range,
                default_value=1e-4
            ),
            ConditionParameter(
                name=ParamNameEnum.EPSILON,
                value_type=ValueTypeEnum.FLOAT64,
                support_disable=False,
                default_value=1e-9,
                visible_on_ui=False
            )
        ],
        supported_target_type=TargetTypeEnum.WEIGHT,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.ACTIVATION_RANGE,
        abbr="AR",
        # Send this condition to MindSpore will use WatchCondition.Condition.activation_range
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.RANGE_START_INCLUSIVE,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range,
                param_type=ParamTypeEnum.SUPPORT_PARAM
            ),
            ConditionParameter(
                name=ParamNameEnum.RANGE_END_INCLUSIVE,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range,
                param_type=ParamTypeEnum.SUPPORT_PARAM
            ),
            ConditionParameter(
                name=ParamNameEnum.RANGE_PERCENTAGE_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_percentage_param_range,
                required_params=[ParamNameEnum.RANGE_START_INCLUSIVE.value, ParamNameEnum.RANGE_END_INCLUSIVE.value]
            ),
            ConditionParameter(
                name=ParamNameEnum.RANGE_PERCENTAGE_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_percentage_param_range,
                required_params=[ParamNameEnum.RANGE_START_INCLUSIVE.value, ParamNameEnum.RANGE_END_INCLUSIVE.value]
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_MIN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_positive_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_MIN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range
            )
        ],
        supported_target_type=TargetTypeEnum.ACTIVATION,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    ),
    Condition(
        condition_id=ConditionIdEnum.TENSOR_RANGE,
        abbr="TR",
        # Send this condition to MindSpore will use WatchCondition.Condition.tensor_range
        optimize_phase=OptimizePhaseEnum.TENSOR_CHECK,
        parameters=[
            ConditionParameter(
                name=ParamNameEnum.RANGE_START_INCLUSIVE,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range,
                param_type=ParamTypeEnum.SUPPORT_PARAM
            ),
            ConditionParameter(
                name=ParamNameEnum.RANGE_END_INCLUSIVE,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_normal_param_range,
                param_type=ParamTypeEnum.SUPPORT_PARAM
            ),
            ConditionParameter(
                name=ParamNameEnum.RANGE_PERCENTAGE_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_percentage_param_range,
                required_params=[ParamNameEnum.RANGE_START_INCLUSIVE.value, ParamNameEnum.RANGE_END_INCLUSIVE.value]
            ),
            ConditionParameter(
                name=ParamNameEnum.RANGE_PERCENTAGE_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_percentage_param_range,
                required_params=[ParamNameEnum.RANGE_START_INCLUSIVE.value, ParamNameEnum.RANGE_END_INCLUSIVE.value]
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_MIN_LT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_positive_param_range
            ),
            ConditionParameter(
                name=ParamNameEnum.MAX_MIN_GT,
                value_type=ValueTypeEnum.FLOAT64,
                valid_test_func=check_abs_param_range
            )
        ],
        supported_target_type=TargetTypeEnum.TENSOR,
        supported_platforms=(PlatformEnum.ASCEND, PlatformEnum.GPU),
        minimum_debugger_capability=(1, 1)
    )
]



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to define the MindSpore graph."""
from mindinsight.datavisual.common.enums import PluginNameEnum
from mindinsight.datavisual.common.log import logger
from mindinsight.datavisual.data_transform.graph.graph import EdgeTypeEnum, Graph, check_invalid_character
from mindinsight.datavisual.data_transform.graph.node import Node
from mindinsight.datavisual.data_transform.graph.node_tree import NodeTree
from mindinsight.datavisual.proto_files.mindinsight_anf_ir_pb2 import DataType
from mindinsight.domain.graph.base import NodeTypeEnum, DebuggerSource


class MSGraph(Graph):
    """The object describes the MindSpore graph, and it is defined in the anf_ir proto file."""

    def _parse_data(self, proto_data):
        """
        The proto data is parsed and all nodes are stored in the specified structure.

        Args:
            proto_data (anf_ir_pb2.GraphProto): Refer to anf_ir_pb2.GraphProto object.
        """
        logger.info("Start to parse graph proto data.")

        self._parse_op_nodes(proto_data.node)
        self._parse_parameters(proto_data.parameters)
        self._parse_consts(proto_data.const_vals)

        self._update_input_after_create_node()
        self._update_output_after_create_node()

        logger.info("Parse proto data end, normal node count(only contain op node, "
                    "parameter, const): %s.", self.normal_node_count)

    def _parse_op_nodes(self, node_protos):
        """
        Parse `anf_ir_pb2.NodeProto` object, and create a normal node.

        Args:
            node_protos (list[anf_ir_pb2.NodeProto]): Refer to anf_ir_pb2.NodeProto.
        """
        logger.debug("Start to parse op nodes from proto.")
        for topological_index, node_proto in enumerate(node_protos):
            if not node_proto.name:
                logger.warning("Finding a node with an empty name will not save it.")
                continue

            self._parse_op_node(topological_index, node_proto)

    def _parse_op_node(self, topological_index, node_proto):
        """Parse `anf_ir_pb2.NodeProto` object, and create a normal node."""
        name = node_proto.name.split('/')[-1]
        node_id = name.split('op')[-1]
        name = f'{node_proto.op_type}-op{node_id}'
        node_name = Node.create_node_name(node_proto.scope, name)

        if node_proto.full_name:
            node_name = node_proto.full_name

        if node_proto.full_name and any(
                node_proto.full_name.lower().endswith(f'[:{plugin.value.lower()}]') for plugin in PluginNameEnum):
            node_name = Node.create_node_name(scope=node_proto.scope,
                                              base_name=f'{node_proto.op_type}-op{node_proto.name}')

        # The Graphviz plug-in that the UI USES can't handle these special characters.
        check_invalid_character(node_name)

        node = Node(name=node_name, node_id=node_id, topological_index=topological_index)
        node.full_name = node_proto.full_name
        node.type = node_proto.op_type
        if getattr(node_proto, 'source_address', None):
            node.stack = DebuggerSource.build_stack_from_source_address(node_proto.source_address)
        self._parse_attributes(node_proto.attribute, node)
        self._parse_inputs(node_proto.input, node)

        node.output_i = node_proto.output_i
        node.scope = node_proto.scope
        node.output_shape = self._get_shape_by_parse_type_proto(node_proto.output_type)
        node.output_nums = len(node.output_shape)
        node.output_data_type = self._get_data_type_by_parse_type_proto(node_proto.output_type, node)

        self._cache_node(node)

    def _parse_parameters(self, parameter_protos):
        """
        Parse `anf_ir_pb2.ParameterProto` object, and create a parameter node.

        Args:
            parameter_protos (list[anf_ir_pb2.ParameterProto]): Refer to anf_ir_pb2.ParameterProto.
        """
        logger.debug("Start to parse parameters from proto.")
        for parameter in parameter_protos:
            if not parameter.name:
                logger.warning("Finding a parameter with an empty name will not save it.")
                continue
            check_invalid_character(parameter.name)
            node = Node(name=parameter.name, node_id=parameter.name)
            node.type = NodeTypeEnum.PARAMETER.value
            node.output_shape = self._get_shape_by_parse_type_proto(parameter.type)
            node.output_nums = len(node.output_shape)
            node.output_data_type = self._get_data_type_by_parse_type_proto(parameter.type, node)
            attr = dict(
                type=self._get_data_type_by_parse_type_proto(parameter.type, node),
                shape=str(self._get_shape_by_parse_type_proto(parameter.type))
            )
            node.add_attr(attr)

            self._cache_node(node)
            logger.debug("Foreach graph proto parameters, node id: %s, node name: %s, "
                         "node def name: %s", node.node_id, node.name, parameter.name)

    def _parse_consts(self, consts):
        """
        Parse `anf_ir_pb2.NameValueProto` object, and create a const node.

        Args:
            consts (list[anf_ir_pb2.NameValueProto]): Refer to `anf_ir_pb2.NameValueProto` object.
        """
        logger.debug("Start to parse consts from proto.")
        for const in consts:
            if not const.key:
                logger.warning("Finding a const with an empty key will not save it.")
                continue
            check_invalid_character(const.key)
            node = Node(name=const.key, node_id=const.key, full_name=const.full_name)
            node.type = NodeTypeEnum.CONST.value
            if const.value.ByteSize() > self.MAX_NODE_ATTRIBUTE_VALUE_BYTES:
                node.add_attr({const.key: 'dtype: ' + DataType.Name(const.value.dtype)})
            else:
                node.add_attr({const.key: str(const.value)})

            if const.value.dtype == DataType.DT_TENSOR:
                shape = list(const.value.tensor_val.dims)
                node.output_shape.append(shape)
                if const.value.tensor_val.HasField('data_type'):
                    node.elem_types.append(DataType.Name(const.value.tensor_val.data_type))
            else:
                node.elem_types.append(DataType.Name(const.value.dtype))
                # dim is zero
                node.output_shape.append([])

            node.output_nums = len(node.output_shape)

            self._cache_node(node)

    def _get_shape_by_parse_type_proto(self, type_proto):
        """
        Parse proto's `message TypeProto` to get shape information.

        Args:
            type_proto (anf_ir_pb2.TypeProto): Refer to anf_ir_pb2.TypeProto.

        Returns:
            list, a list of shape.
        """
        shapes = []
        if type_proto.HasField('data_type'):
            if type_proto.data_type != DataType.DT_TENSOR and \
                    type_proto.data_type != DataType.DT_TUPLE:
                # Append an empty list as a placeholder
                # for the convenience of output number calculation.
                shapes.append([])
                return shapes
        if type_proto.HasField('tensor_type'):
            tensor_type = type_proto.tensor_type
            tensor_shape_proto = tensor_type.shape
            shape = [dim.size for dim in tensor_shape_proto.dim]
            shapes.append(shape)
        if type_proto.HasField('sequence_type'):
            for elem_type in type_proto.sequence_type.elem_types:
                shapes.extend(self._get_shape_by_parse_type_proto(elem_type))
        return shapes

    def _get_data_type_by_parse_type_proto(self, type_proto, node):
        """
        Get data type by parse type proto object.

        The name of the DataType, refer to `anf_ir_pb2.DataType` object.
        If data type is tensor or tuple, the data name we return is `data_type[element_type, element_type]`.

        Args:
            type_proto (anf_ir_pb2.TypeProto): Refer to anf_ir_pb2.TypeProto.

        Returns:
            str, the data type.

        """
        data_type_name = self._get_data_type_name_by_value(type_proto, type_proto.data_type, field_name='data_type')
        if type_proto.data_type == DataType.DT_TENSOR:
            tensor_type_proto = type_proto.tensor_type
            value = type_proto.tensor_type.elem_type
            elem_type_name = self._get_data_type_name_by_value(tensor_type_proto, value, field_name='elem_type')
            node.elem_types.append(elem_type_name)
            return f'{data_type_name}[{elem_type_name}]'

        if type_proto.data_type == DataType.DT_TUPLE:
            data_types = []
            for elem_type in type_proto.sequence_type.elem_types:
                data_types.append(self._get_data_type_by_parse_type_proto(elem_type, node))
            return f'{data_type_name}{str(data_types)}'

        node.elem_types.append(data_type_name)

        return data_type_name

    def get_nodes(self, searched_node_list):
        """
        Get node tree by a searched_node_list.

        Args:
            searched_node_list (list[Node]): A list of nodes that
                matches the given search pattern.

        Returns:
            A list of dict including the searched nodes.
            [{
                "name": "Default",
                "type": "name_scope",
                "nodes": [{
                    "name": "Default/Conv2D1",
                    "type": "name_scope",
                    "nodes": [{
                        ...
                    }]
                }]
            },
            {
                "name": "Gradients",
                "type": "name_scope",
                "nodes": [{
                    "name": "Gradients/Default",
                    "type": "name_scope",
                    "nodes": [{
                        ...
                }]
            }]
        """
        # save the node in the NodeTree
        root = NodeTree()
        for node in searched_node_list:
            self._build_node_tree(root, node.name, node.type, node.is_dynamic_shape_node)

        # get the searched nodes in the NodeTree and reorganize them
        searched_list = []
        self._traverse_node_tree(root, searched_list)

        return searched_list

    def search_leaf_nodes_by_pattern(self, pattern, scope_pattern=False):
        """
        Search leaf node by a given pattern.

        Args:
            pattern (Union[str, None]): The pattern of the node to search,
                if None, return all node names.
            scope_pattern (bool): If true, return the children nodes of the scope. Default: False.

        Returns:
            list[Node], a list of nodes.
        """
        is_match = lambda x, y: x.lower().startswith(y) if scope_pattern else y in x.lower()
        if pattern is not None:
            pattern = pattern.lower()
            searched_nodes = [
                node for name, node in self._leaf_nodes.items()
                if is_match(name, pattern)
            ]
        else:
            searched_nodes = [node for node in self._leaf_nodes.values()]
        return searched_nodes

    def search_nodes_by_pattern(self, pattern):
        """
        Search node by a given pattern.

        Search node which pattern is the part of the last node. Example: pattern=ops, node1=default/ops,
        node2=default/ops/weight, so node2 will be ignore and only node1 will be return.

        Args:
            pattern (Union[str, None]): The pattern of the node to search.

        Returns:
            list[Node], a list of nodes.
        """
        searched_nodes = []
        if pattern and pattern != '/':
            pattern = pattern.lower()
            for name, node in self._normal_node_map.items():
                name = name.lower()
                pattern_index = name.rfind(pattern)
                if pattern_index >= 0 and name.find('/', pattern_index + len(pattern)) == -1:
                    searched_nodes.append(node)
        return searched_nodes

    def _build_node_tree(self, root, node_name, node_type, is_dynamic_shape_node):
        """
        Build node tree.

        Args:
            root (NodeTree): Root node of node tree.
            node_name (str): Node name.
            node_type (str): Node type.
        """
        scope_names = node_name.split('/')
        cur_node = root
        full_name = ""
        for scope_name in scope_names[:-1]:
            full_name = '/'.join([full_name, scope_name]) if full_name else scope_name
            scope_node = self._get_normal_node(node_name=full_name)
            sub_node = cur_node.get(scope_name)
            if not sub_node:
                sub_node = cur_node.add(scope_name, scope_node.type, is_dynamic_shape_node)
            cur_node = sub_node
        cur_node.add(scope_names[-1], node_type, is_dynamic_shape_node)

    def _traverse_node_tree(self, cur_node, search_node_list):
        """Traverse the node tree and construct the searched nodes list."""
        for _, sub_node in cur_node.get_children():
            sub_nodes = []
            self._traverse_node_tree(sub_node, sub_nodes)
            sub_node_dict = {
                'name': sub_node.node_name,
                'type': sub_node.node_type,
                'is_dynamic_shape_node': sub_node.is_dynamic_shape_node,
                'nodes': sub_nodes
            }
            search_node_list.append(sub_node_dict)

    def _parse_inputs(self, input_protos, node):
        """
        Parse `anf_ir_pb2.InputProto` object.

        Args:
            input_protos (list[anf_ir_pb2.InputProto]): Refer to `anf_ir_pb2.InputProto` object.
            node (Node): Refer to `Node` object, it is used to log message and update input.
        """
        for input_proto in input_protos:
            if not input_proto.name:
                logger.warning("The name in input proto of node(%s) is empty, will ignore.", node.name)
                continue

            if "/" in input_proto.name:
                input_proto.name = input_proto.name.split("op")[-1]

            edge_type = EdgeTypeEnum.DATA.value if not input_proto.type else EdgeTypeEnum.CONTROL.value

            # Notice:
            # 1. The name in the input proto is the node id of the Node object.
            # 2. In the current step, the shape of source node cannot be obtained,
            #    so it is set to empty list by default, and the next step will update it.
            # 3. Same with scope, set the default value first.
            input_attr = {
                "shape": [],
                "edge_type": edge_type,
                "independent_layout": False,
                'data_type': ''
            }

            node.add_inputs(src_name=input_proto.name, input_attr=input_attr)

    def _parse_attributes(self, attributes, node):
        """
        Parse `anf_ir_pb2.AttributeProto` object., and Filters large attribute values.

        Args:
            attributes (list[anf_ir_pb2.AttributeProto]): Refer to `anf_ir_pb2.AttributeProto` object.
            node (Node): Refer to `Node` object, it is used to log message and update attr.
        """
        for attr in attributes:
            if attr.value.ByteSize() > self.MAX_NODE_ATTRIBUTE_VALUE_BYTES:
                message = f"The attribute value of node({node.name}) " \
                          f"is over {self.MAX_NODE_ATTRIBUTE_VALUE_BYTES} Bytes, will ignore."
                logger.warning(message)
                continue
            if attr.name in ('input_is_dynamic_shape', 'output_is_dynamic_shape') and not \
                    node.is_dynamic_shape_node and attr.value.bool_val:
                node.is_dynamic_shape_node = True
            node.add_attr({attr.name: str(attr.value)})

    def _update_input_after_create_node(self):
        """Update the input of node after create node."""
        for node in self._normal_node_map.values():
            for src_node_id, input_attr in dict(node.inputs).items():
                node.delete_inputs(src_node_id)
                if not self._is_node_exist(node_id=src_node_id):
                    message = f"The input node could not be found by node id({src_node_id}) " \
                              f"while updating the input of the node({node})"
                    logger.warning(message)

                    continue

                src_node = self._get_normal_node(node_id=src_node_id)
                input_attr['shape'] = src_node.output_shape
                input_attr['data_type'] = src_node.output_data_type
                node.add_inputs(src_name=src_node.name, input_attr=input_attr)

    def _update_output_after_create_node(self):
        """Update the output of node after create node."""
        # Constants and parameter should not exist for input and output.
        filtered_node = {NodeTypeEnum.CONST.value, NodeTypeEnum.PARAMETER.value}
        for node in self._normal_node_map.values():
            for src_name, input_attr in node.inputs.items():
                src_node = self._get_normal_node(node_name=src_name)
                if src_node.type in filtered_node:
                    continue

                src_node.add_outputs(node.name, input_attr)

    @staticmethod
    def _get_data_type_name_by_value(data_type, value, field_name='data_type'):
        """Get the data type name by the enum value, data_type refer to `DataType` object."""
        return data_type.DESCRIPTOR.fields_by_name[field_name].enum_type.values_by_number[value].name



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""The proposer base class."""
from abc import ABC, abstractmethod

from marshmallow import ValidationError

from mindinsight.profiler.common.log import logger
from mindinsight.profiler.analyser.analyser_factory import AnalyserFactory
from mindinsight.utils.exceptions import MindInsightException
from mindinsight.profiler.common.validator.validate_path import \
    validate_and_normalize_path
from mindinsight.profiler.common.exceptions.exceptions import ProfilerPathErrorException


class Proposer(ABC):
    """The proposer base class."""

    def __init__(self, profiling_path, rank_id):
        self.profiling_path = self._normalize_profiling_dir(profiling_path)
        self.rank_id = rank_id

    def get_analyser_result(self, analyser_type, condition=None):
        logger.debug("The Proposer 'analyser_type' is %s, 'options' is %s", str(analyser_type), str(condition))
        analyser_result = {}
        try:
            analyser = AnalyserFactory.instance().get_analyser(analyser_type, self.profiling_path, self.rank_id)
            analyser_result = analyser.query(condition)
            logger.debug("The 'analyser_result' is %s, the 'condition' is %s.", str(analyser_result), str(condition))
        except MindInsightException as e:
            logger.warning(e)
        return analyser_result

    @abstractmethod
    def analyze(self, options=None):
        """analysis and get proposal."""
        raise NotImplementedError("Must define analyze function to inherit Class Propose")

    @staticmethod
    def _normalize_profiling_dir(profiling_dir):
        """
        Normalize the profiling dir.

        Args:
              profiling_dir (str): The directory where the parsed profiling files
                 are located.

        Returns:
              str, the normalized profiling dir.
        """
        try:
            normalized_profiling_dir = validate_and_normalize_path(
                profiling_dir, 'profiler'
            )
        except ValidationError:
            raise ProfilerPathErrorException('The profiling dir is invalid.')
        return normalized_profiling_dir



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Implement the session manager."""
import os
import threading

import _thread

from mindinsight.conf import settings
from mindinsight.debugger.common.log import LOGGER as logger
from mindinsight.debugger.common.exceptions.exceptions import DebuggerSessionNumOverBoundError, \
    DebuggerSessionNotFoundError, DebuggerOnlineSessionUnavailable
from mindinsight.debugger.debugger_services.debugger_server_factory import DebuggerServerContext
from mindinsight.debugger.debugger_session import DebuggerSession


class SessionManager:
    """The server manager of debugger."""

    ONLINE_TYPE = "ONLINE"
    MAX_OFFLINE_SESSION_NUM = settings.MAX_OFFLINE_DEBUGGER_SESSION_NUM
    ONLINE_SESSION_ID = "0"
    _instance = None
    _cls_lock = threading.Lock()

    def __init__(self):
        self.train_jobs = {}
        self.sessions = {}
        # The offline session id is start from 1, and the online session id is 0.
        self._next_session_id = 1
        self._lock = threading.Lock()
        self._exiting = False
        enable_debugger = settings.ENABLE_DEBUGGER if hasattr(settings, 'ENABLE_DEBUGGER') else False
        if enable_debugger:
            self._create_online_session()

    @classmethod
    def get_instance(cls):
        """Get the singleton instance."""
        with cls._cls_lock:
            if cls._instance is None:
                cls._instance = SessionManager()
            return cls._instance

    def exit(self):
        """Called when the gunicorn worker process is exiting."""
        with self._lock:
            logger.info("Start to exit sessions.")
            self._exiting = True
            for session in self.sessions.values():
                session.stop()
        logger.info("Sessions exited.")

    def get_session(self, session_id):
        """
        Get session by session id or get all session info.

        Args:
            session_id (Union[None, str]): The id of session.

        Returns:
            DebuggerSession, debugger session object.
        """
        with self._lock:
            if session_id in self.sessions:
                return self.sessions.get(session_id)

            logger.error('Debugger session %s is not found.', session_id)
            raise DebuggerSessionNotFoundError("{}".format(session_id))

    def _create_online_session(self):
        """Create online session."""
        with self._lock:
            context = DebuggerServerContext(dbg_mode='online')
            online_session = DebuggerSession(context)
            online_session.start()
            self.sessions[self.ONLINE_SESSION_ID] = online_session

    def _create_offline_session(self, train_job):
        """Create offline session."""
        self._check_session_num()
        if not isinstance(train_job, str):
            logger.error('The train job path should be string.')
            raise ValueError("The train job path should be string.")
        summary_base_dir = settings.SUMMARY_BASE_DIR
        whole_path = os.path.join(summary_base_dir, train_job)
        normalized_path = validate_and_normalize_path(whole_path)
        context = DebuggerServerContext(dbg_mode='offline', train_job=train_job, dbg_dir=normalized_path)
        session = DebuggerSession(context)
        session.start()
        session_id = str(self._next_session_id)
        self.sessions[session_id] = session
        self.train_jobs[train_job] = session_id
        self._next_session_id += 1
        return session_id

    def create_session(self, session_type, train_job=None):
        """
        Create the session by the train job info or session type if the session doesn't exist.

        Args:
            session_type (str): The session_type.
            train_job (str): The train job info.

        Returns:
            str, session id.
        """
        with self._lock:
            if self._exiting:
                logger.info(
                    "System is exiting, will terminate the thread.")
                _thread.exit()

            if session_type == self.ONLINE_TYPE:
                if self.ONLINE_SESSION_ID not in self.sessions:
                    logger.error(
                        'Online session is unavailable, set --enable-debugger as true/1 to enable debugger '
                        'when start Mindinsight server.')
                    raise DebuggerOnlineSessionUnavailable()
                return self.ONLINE_SESSION_ID

            if train_job in self.train_jobs:
                return self.train_jobs.get(train_job)

            return self._create_offline_session(train_job)

    def delete_session(self, session_id):
        """Delete session by session id."""
        with self._lock:
            if session_id == self.ONLINE_SESSION_ID:
                logger.error('Online session can not be deleted.')
                raise ValueError("Online session can not be delete.")

            if session_id not in self.sessions:
                logger.error('Debugger session %s is not found', session_id)
                raise DebuggerSessionNotFoundError("session id {}".format(session_id))

            session = self.sessions.get(session_id)
            session.stop()
            self.sessions.pop(session_id)
            self.train_jobs.pop(session.train_job)
            return

    def get_train_jobs(self):
        """Get all train jobs."""
        return {"train_jobs": self.train_jobs}

    def _check_session_num(self):
        """Check the amount of sessions."""
        session_limitation = self.MAX_OFFLINE_SESSION_NUM
        if self.ONLINE_SESSION_ID in self.sessions:
            session_limitation += 1
        if len(self.sessions) >= session_limitation:
            logger.warning('Offline debugger session num %s is reach the limitation %s', len(self.sessions),
                           session_limitation)
            raise DebuggerSessionNumOverBoundError()


def validate_and_normalize_path(path):
    """Validate and normalize_path"""
    if not path:
        logger.error('The whole path of dump directory is None.')
        raise ValueError("The whole path of dump directory is None.")

    path_str = str(path)

    if not path_str.startswith("/"):
        logger.error('The whole path of dump directory is not start with \'/\'')
        raise ValueError("The whole path of dump directory is not start with \'/\'")

    try:
        normalized_path = os.path.realpath(path)
    except ValueError:
        logger.error('The whole path of dump directory is invalid.')
        raise ValueError("The whole path of dump directory is invalid.")

    return normalized_path



# Copyright 2020-2022 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""
Profiler util.

This module provides the utils.
"""
import json
import os
import re
import glob

from mindinsight import __version__ as mi_version
from mindinsight.datavisual.utils.tools import to_int
from mindinsight.profiler.common.exceptions.exceptions import ProfilerDirNotFoundException
from mindinsight.profiler.common.exceptions.exceptions import ProfilerInfoFieldMissingException
from mindinsight.datavisual.common.exceptions import TrainJobNotExistError
from mindinsight.common.util import version_match

# one sys count takes 10 ns, 1 ms has 100000 system count
PER_MS_SYSCNT = 100000


def analyse_device_list_from_profiler_dir(profiler_dir):
    """
    Analyse device list from profiler dir.

    Args:
        profiler_dir (str): The profiler data dir.

    Returns:
        list, the device_id list.
    """
    profiler_file_prefix = ("ascend_timeline_display*", "output_op_compute_time*")
    gpu_profiler_file_prefix = ("gpu_op_detail_info*", "gpu_activity_data*", "gpu_op_type_info*")
    cpu_profiler_file_prefix = ("cpu_op_detail_info*", "cpu_op_type_info*")

    def get_device_list_and_profiler_mode():
        profiler_info_list = glob.glob(fr'{profiler_dir}/profiler_info_*.json')
        device_id_list = set()
        profiler_mode = ''
        for profiler_info_file in profiler_info_list:
            if not profiler_mode:
                with os.fdopen(os.open(profiler_info_file, os.O_RDONLY, 0o400), 'r') as fr:
                    json_data = json.load(fr)
                profiler_mode = json_data.get('context_mode', '')
            device_id = profiler_info_file.split('profiler_info_')[-1].split('.')[0]
            device_id_list.add(device_id)
        return sorted(list(device_id_list)), profiler_mode

    device_id_list, profiler_mode = get_device_list_and_profiler_mode()

    def get_profiler_type(file_prefix):
        for file_name in file_prefix:
            if glob.glob(fr'{profiler_dir}/{file_name}'):
                return True
        return False

    if get_profiler_type(profiler_file_prefix):
        return device_id_list, 'ascend', profiler_mode

    if get_profiler_type(gpu_profiler_file_prefix):
        return device_id_list, 'gpu', profiler_mode

    if get_profiler_type(cpu_profiler_file_prefix):
        return device_id_list, 'cpu', profiler_mode

    return device_id_list, '', profiler_mode


def query_latest_trace_time_file(profiler_dir, device_id=0):
    """
    Query the latest trace time file.

    Args:
        profiler_dir (str): The profiler directory.
        device_id (int): The id of device.

    Returns:
        str, the latest trace time file path.
    """
    files = os.listdir(profiler_dir)
    target_file = f'step_trace_raw_{device_id}_detail_time.csv'
    try:
        latest_file = max(
            filter(
                lambda file: file == target_file,
                files
            ),
            key=lambda file: os.stat(os.path.join(profiler_dir, file)).st_mtime
        )
    except ValueError:
        return None
    return os.path.join(profiler_dir, latest_file)


def query_profiling_graph_mode(profiler_dir):
    graph_mode = 'static'
    dynamic_shape_file_prefix = "dynamic_shape_info"
    all_file = os.listdir(profiler_dir)
    for file in all_file:
        if file.startswith(dynamic_shape_file_prefix):
            graph_mode = 'dynamic'
            return graph_mode
    return graph_mode


def query_step_trace_file(profiler_dir):
    """
    Query for all step trace file.

    Args:
        profiler_dir (str): The directory that contains all step trace files.

    Returns:
        str, the file path of step trace time.
    """
    files = os.listdir(profiler_dir)
    training_trace_file = list(
        filter(
            lambda file: file.startswith('training_trace') and not file.endswith('.done'),
            files
        )
    )
    if training_trace_file:
        return os.path.join(profiler_dir, training_trace_file[0])
    return None


def get_summary_for_step_trace(average_info, header):
    """The property of summary info."""
    if not average_info or not header:
        return {}
    total_time = get_field_value(average_info, 'total', header)
    iteration_interval = get_field_value(average_info, 'iteration_interval',
                                         header)
    summary_part = {
        'total_time': total_time,
        'iteration_interval': iteration_interval,
        'iteration_interval_percent': calculate_percent(iteration_interval, total_time),
    }
    # training scene data for ui display
    if 'fp_and_bp' in header:
        fp_and_bp = get_field_value(average_info, 'fp_and_bp', header)
        tail = get_field_value(average_info, 'tail', header)
        summary = {
            'fp_and_bp': fp_and_bp,
            'fp_and_bp_percent': calculate_percent(fp_and_bp, total_time),
            'tail': tail,
            'tail_percent': calculate_percent(tail, total_time)
        }
    # inference scene data for ui display
    else:
        fp = get_field_value(average_info, 'fp', header)
        summary = {
            'fp': fp,
            'fp_percent': calculate_percent(fp, total_time)
        }
    summary.update(summary_part)
    return summary


def calculate_percent(partial, total):
    """Calculate percent value."""
    if total:
        percent = round(partial / total * 100, 2)
    else:
        percent = 0
    return f'{percent}%'


def to_millisecond(sys_count, limit=4):
    """Translate system count to millisecond."""
    return round(sys_count / PER_MS_SYSCNT, limit)


def get_field_value(row_info, field_name, header, time_type='realtime'):
    """
    Extract basic info through row_info.

    Args:
        row_info (list): The list of data info in one row.
        field_name (str): The name in header.
        header (list[str]): The list of field names.
        time_type (str): The type of value, `realtime` or `systime`. Default: `realtime`.

    Returns:
        dict, step trace info in dict format.
    """
    field_index = header.index(field_name)
    value = row_info[field_index]
    value = to_int(value, field_name)
    if time_type == 'realtime':
        value = to_millisecond(value)

    return value


def get_options(options):
    """Get options."""
    if options is None:
        options = {}
    return options


def check_train_job_and_profiler_dir(profiler_dir_abs):
    """ check the existence of train_job and profiler dir """
    train_job_dir_abs = os.path.abspath(os.path.join(profiler_dir_abs, '..'))
    if not os.path.exists(train_job_dir_abs):
        raise TrainJobNotExistError(error_detail=train_job_dir_abs)
    if not os.path.exists(profiler_dir_abs):
        raise ProfilerDirNotFoundException(msg=profiler_dir_abs)


def get_profile_data_version(profiler_dir):
    """ get the profiler data version """

    profile_info_pattern = re.compile(r"profiler_info_(\d+).json")
    profile_info_file = None
    for f_name in os.listdir(profiler_dir):
        re_match = re.match(profile_info_pattern, f_name)
        if re_match:
            profile_info_file = re_match.group()
            break
    if profile_info_file:
        full_path = os.path.join(profiler_dir, profile_info_file)
        with open(full_path, 'r') as fr:
            data = json.load(fr)
        ms_data_version = data.get('ms_version', '')
        if ms_data_version:
            data_match_stat = version_match(ms_data_version, mi_version)
            stat_reply = True
            if data_match_stat >= 1:
                stat_reply = False
            result_data = {'state': stat_reply,
                           'ms': ms_data_version,
                           'mi': mi_version}
            return result_data
    return {}


def get_all_export_flag(profiler_dir):
    """ get the flag what means whether support all-export mode. """

    profile_info_pattern = re.compile(r"profiler_info_(\d+).json")
    profile_info_file = None
    for f_name in os.listdir(profiler_dir):
        re_match = re.match(profile_info_pattern, f_name)
        if re_match:
            profile_info_file = re_match.group()
            break
    if profile_info_file:
        full_path = os.path.join(profiler_dir, profile_info_file)
        with open(full_path, 'r') as fr:
            data = json.load(fr)
        return data.get('all_export', False)

    return False


def get_parallel_message(profiler_dir):
    """get the parallel message"""

    profile_info_pattern = re.compile(r"profiler_info_(\d+).json")
    profile_info_file = None
    for f_name in os.listdir(profiler_dir):
        re_match = re.match(profile_info_pattern, f_name)
        if re_match:
            profile_info_file = re_match.group()
            break
    if profile_info_file:
        full_path = os.path.join(profiler_dir, profile_info_file)
        with open(full_path, 'r') as fr:
            data = json.load(fr)
        parallel_mode = data.get('parallel_mode')
        stage_num = data.get('stage_num')
        rank_size = data.get('rank_size')
        if not parallel_mode or not stage_num or not rank_size:
            msg = f"file: {full_path}, fields: parallel_mode, stage_num, rank_size"
            raise ProfilerInfoFieldMissingException(msg=msg)
        return {'parallel_mode': parallel_mode, 'stage_num': stage_num, 'rank_size': rank_size}
    return {}



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to identify the type of the node."""
import sys

from mindinsight.domain.graph.base import NodeTypeEnum
from mindinsight.debugger.common.exceptions.exceptions import DebuggerParamValueError

_ACTIVATIONS = [
    'elu',
    'fastgelu',
    'gelu',
    'hsigmoid',
    'hswish',
    'leakyrelu',
    'logsigmoid',
    'logsoftmax',
    'prelu',
    'relu',
    'relu6',
    'reluv2',
    'sigmoid',
    'softmax',
    'tanh'
]


class NodeTypeIdentifier:
    """Node type identifier."""

    def __init__(self, node_type):
        self.identify_func = self.get_identify_func(node_type)

    @staticmethod
    def get_identify_func(node_type):
        """
        Get the identify function in this module.

        Args:
            node_type (str): The node type.

        Returns:
            function, the identify function.
        """
        # the name of the identity function should start with 'is_' and end with '_node'
        target_name = 'is_' + node_type + '_node'
        cur_module = sys.modules[__name__]
        for sub_module in dir(cur_module):
            # the rule to get the identify function
            if sub_module == target_name:
                return getattr(cur_module, sub_module)
        raise DebuggerParamValueError("Invalid identify type.")

    def is_match(self, *args, **kwargs):
        """Check if the input match the idenfity function."""
        return self.identify_func(*args, **kwargs)


def is_parameter_node(node):
    """
    Check if the node is weight type.

    Args:
        node (Node): The node object.

    Returns:
        bool, if the node is weight type.
    """
    return bool(node.type == NodeTypeEnum.PARAMETER.value)


def is_weight_node(node):
    """
    Check if the node is weight type.

    Args:
        node (Node): The node object.

    Returns:
        bool, if the node is weight type.
    """
    if node.type == NodeTypeEnum.PARAMETER.value:
        full_name = node.full_name.lower()
        weight_flag = False
        if full_name.endswith('.weight') or full_name.endswith('.bias'):
            weight_flag = True
        if weight_flag and 'optimizer-' not in full_name and not full_name.startswith('gradients/'):
            return True
    return False


def is_activation_node(node, condition=None):
    """
    Check if the node is activation type.

    Args:
        node (Node): The node object.
        condition (dict): Filter condition.

            - activation_func (Union[str, list[str]): The target functions.

    Returns:
        bool, if the node is activation type.
    """
    activation_funcs = condition.get('activation_func') if condition else _ACTIVATIONS
    if not activation_funcs:
        activation_funcs = _ACTIVATIONS
    if not isinstance(activation_funcs, list):
        activation_funcs = [activation_funcs]

    if not is_gradient_node(node):
        node_type = node.type
        for activation_name in activation_funcs:
            if node_type.lower() == activation_name:
                return True
    return False


def is_gradient_node(node):
    """
    Check if the node is gradient type.

    Args:
        node (Node): The node object.

    Returns:
        bool, if the node is gradient type.
    """
    full_name = node.full_name.lower()
    if (full_name == 'gradients' or full_name.startswith('gradients/')) and \
            node.type not in [NodeTypeEnum.PARAMETER.value, NodeTypeEnum.CONST.value]:
        return True
    return False


def is_tensor_node(node):
    """
    Check if the node is tensor type.

    Args:
        node (Node): The node object.

    Returns:
        bool, if the node is tensor type.
    """
    if node is not None:
        return True
    return False



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Hyper config."""
import json
import os

from mindinsight.optimizer.common.constants import HYPER_CONFIG_ENV_NAME, HYPER_CONFIG_LEN_LIMIT
from mindinsight.optimizer.common.exceptions import HyperConfigEnvError, HyperConfigError


class AttributeDict(dict):
    """A dict can be accessed by attribute."""
    def __init__(self, d=None):
        super().__init__()
        if d is not None:
            for k, v in d.items():
                self[k] = v

    def __key(self, key):
        """Get key."""
        return "" if key is None else key

    def __setattr__(self, key, value):
        """Set attribute for object."""
        self[self.__key(key)] = value

    def __getattr__(self, key):
        """
        Get attribute value according by attribute name.

        Args:
            key (str): attribute name.

        Returns:
            Any, attribute value.

        Raises:
            AttributeError: If the key does not exists, will raise Exception.

        """
        value = self.get(self.__key(key))
        if value is None:
            raise AttributeError("The attribute %r is not exist." % key)
        return value

    def __getitem__(self, key):
        """Get attribute value according by attribute name."""
        value = super().get(self.__key(key))
        if value is None:
            raise AttributeError("The attribute %r is not exist." % key)
        return value

    def __setitem__(self, key, value):
        """Set attribute for object."""
        return super().__setitem__(self.__key(key), value)


class HyperConfig:
    """
    Hyper config.
    1. Init HyperConfig.
    2. Get suggested params and summary_dir.
    3. Record by SummaryCollector with summary_dir.

    Examples:
        >>> hyper_config = HyperConfig()
        >>> params = hyper_config.params
        >>> learning_rate = params.learning_rate
        >>> batch_size = params.batch_size

        >>> summary_dir = hyper_config.summary_dir
        >>> summary_cb = SummaryCollector(summary_dir)
    """
    def __init__(self):
        self._init_validate_hyper_config()

    def _init_validate_hyper_config(self):
        """Init and validate hyper config."""
        hyper_config = os.environ.get(HYPER_CONFIG_ENV_NAME)
        if hyper_config is None:
            raise HyperConfigEnvError("Hyper config is not in system environment.")
        if len(hyper_config) > HYPER_CONFIG_LEN_LIMIT:
            raise HyperConfigEnvError("Hyper config is too long. The length limit is %s, the length of "
                                      "hyper_config is %s." % (HYPER_CONFIG_LEN_LIMIT, len(hyper_config)))

        try:
            hyper_config = json.loads(hyper_config)
        except TypeError as exc:
            raise HyperConfigError("Hyper config type error. detail: %s." % str(exc))
        except Exception as exc:
            raise HyperConfigError("Hyper config decode error. detail: %s." % str(exc))

        self._validate_hyper_config(hyper_config)
        self._hyper_config = hyper_config

    def _validate_hyper_config(self, hyper_config):
        """Validate hyper config."""
        for key in ['summary_dir', 'params']:
            if key not in hyper_config:
                raise HyperConfigError("%r must exist in hyper_config." % key)

        # validate summary_dir
        summary_dir = hyper_config.get('summary_dir')
        if not isinstance(summary_dir, str):
            raise HyperConfigError("The 'summary_dir' should be string.")
        hyper_config['summary_dir'] = os.path.realpath(summary_dir)

        # validate params
        params = hyper_config.get('params')
        if not isinstance(params, dict):
            raise HyperConfigError("'params' is not a dict.")
        for key, value in params.items():
            if not isinstance(value, (int, float)):
                raise HyperConfigError("The value of %r is not integer or float." % key)

    @property
    def params(self):
        """Get params."""
        return AttributeDict(self._hyper_config.get('params'))

    @property
    def summary_dir(self):
        """Get train summary dir path."""
        return self._hyper_config.get('summary_dir')

    @property
    def custom_lineage_data(self):
        return self._hyper_config.get('custom_lineage_data')



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""File handler for lineage summary log."""
import os
from mindinsight.lineagemgr.common.exceptions.exceptions import LineageFileHandlerReadError

class FileHandler:
    """
    Summary log file handler.

    Summary log file handler provides Python APIs to manage file IO, including
    read, seek. It is not suitable for very large files.

    Args:
        file_path (str): File path.
    """

    def __init__(self, file_path):
        self._size = os.path.getsize(file_path)
        self._cache = self._read_cache(file_path)
        self._offset = 0

    @property
    def size(self):
        """
        The size of file.

        Returns:
            int, the size of file.
        """
        return self._size

    def _read_cache(self, file_path):
        """
        Read file in cache.

        Args:
            file_path (str): File path.

        Returns:
            bytes, the file content.
        """
        with open(file_path, 'rb') as log_file:
            return log_file.read()

    def seek(self, offset):
        """
        Set the new offset of file.

        Args:
            pos (int): The new offset.
        """
        self._offset = offset

    def tell(self):
        """
        Tell the current offset.

        Returns:
            int, the offset.
        """
        return self._offset

    def read(self, size=-1, offset=None):
        """
        Read bytes from buffer by size.

        Args:
            size (int): Number of bytes to read. If set -1, read the whole file.
                Default: -1.
            offset (int): The start offset to read bytes from. Default: None.

        Returns:
            bytes, the content.
        """
        if offset is None:
            offset = self._offset

        new_offset = offset + size
        if new_offset > self.size:
            raise LineageFileHandlerReadError('The sum of offset and size is out of range.')
        result = self._cache[offset:new_offset]
        self._offset = new_offset

        return result



# Copyright 2020 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""This file is used to define lineage info querier."""
import enum
import functools
import operator

from mindinsight.lineagemgr.common.exceptions.exceptions import LineageQuerierParamException, LineageParamTypeError
from mindinsight.lineagemgr.common.utils import enum_to_list
from mindinsight.lineagemgr.lineage_parser import SuperLineageObj
from mindinsight.lineagemgr.querier.query_model import FIELD_MAPPING


@enum.unique
class ConditionParam(enum.Enum):
    """
    Filtering and sorting field names.

    `LIMIT` represents the number of lineage info per page. `OFFSET` represents
    page number. `SORTED_NAME` means to sort by this field. `SORTED_TYPE` means
    ascending or descending.
    """
    LIMIT = 'limit'
    OFFSET = 'offset'
    SORTED_NAME = 'sorted_name'
    SORTED_TYPE = 'sorted_type'
    LINEAGE_TYPE = 'lineage_type'

    @classmethod
    def is_condition_type(cls, value):
        """
        Judge that the input param is one of field names in the class.

        Args:
            value (str): The input field name.

        Returns:
            bool, `True` if the input field name in the class, else `False`.
        """
        return value in cls._value2member_map_


@enum.unique
class ExpressionType(enum.Enum):
    """
    Filter condition name definition.

    `EQ` means `==`. `LT` means `<`. `GT` means `>`. `LE` means `<=`. `GE` means
    `>=`. `IN` means filter value in the specified list.
    """
    EQ = 'eq'
    LT = 'lt'
    GT = 'gt'
    LE = 'le'
    GE = 'ge'
    IN = 'in'
    NOT_IN = 'not_in'

    @classmethod
    def is_valid_exp(cls, key):
        """
        Judge that the input param is one of filter condition names in the class.

        Args:
            key (str): The input filter condition name.

        Returns:
            bool, `True` if the input filter condition name in the class,
            else `False`.
        """
        return key in cls._value2member_map_

    @classmethod
    def is_match(cls, except_key, except_value, actual_value):
        """
        Determine whether the value meets the expected requirement.

        Args:
            except_key (str): The expression key.
            except_value (Union[str, int, float, list, tuple]): The expected
                value.
            actual_value (Union[str, int, float]): The actual value.

        Returns:
            bool, `True` if the actual value meets the expected requirement,
            else `False`.
        """
        if actual_value is None and except_key in [cls.LT.value, cls.GT.value,
                                                   cls.LE.value, cls.GE.value]:
            return False

        try:
            if except_key == cls.IN.value:
                state = operator.contains(except_value, actual_value)
            elif except_key == cls.NOT_IN.value:
                state = not operator.contains(except_value, actual_value)
            else:
                state = getattr(operator, except_key)(actual_value, except_value)
        except TypeError:
            # actual_value can not compare with except_value
            return False
        return state


@enum.unique
class LineageFilterKey(enum.Enum):
    """Summary lineage information filter key."""
    METRIC = 'metric'
    HYPER_PARAM = 'hyper_parameters'
    ALGORITHM = 'algorithm'
    TRAIN_DATASET = 'train_dataset'
    VALID_DATASET = 'valid_dataset'
    MODEL = 'model'
    DATASET_GRAPH = 'dataset_graph'

    @classmethod
    def is_valid_filter_key(cls, key):
        """
        Judge that the input param is one of field names in the class.

        Args:
            key (str): The input field name.

        Returns:
            bool, `True` if the input field name in the class, else `False`.
        """
        return key in cls._value2member_map_

    @classmethod
    def get_key_list(cls):
        """
        Get the filter key name list.

        Returns:
            list[str], the filter key name list.
        """
        return [member.value for member in cls]


@enum.unique
class LineageType(enum.Enum):
    """Lineage search type."""
    DATASET = 'dataset'
    MODEL = 'model'


class Querier:
    """
    The querier of model lineage information.

    The class provides model lineage information query function. The information
    includes hyper parameters, train dataset, algorithm, model information,
    metric, valid dataset, etc.

    The class also provides search and sorting capabilities about model lineage
    information. You can search and sort by the specified condition.
    The condition explain in `ConditionParam` and `ExpressionType` class.
    See the method `filter_summary_lineage` for supported fields.

    Args:
        super_lineage_objs (dict): A dict of <summary_dir, SuperLineageObject>.

    Raises:
        LineageParamTypeError: If the input parameter type is invalid.
        LineageQuerierParamException: If the input parameter value is invalid.
        LineageSummaryParseException: If all summary logs parsing failed.
    """
    def __init__(self, super_lineage_objs):
        self._super_lineage_objs = self._check_objs(super_lineage_objs)

    def _check_objs(self, super_lineage_objs):
        if super_lineage_objs is None:
            raise LineageQuerierParamException(
                'querier_init_param', 'The querier init param is empty.'
            )
        if not isinstance(super_lineage_objs, dict):
            raise LineageParamTypeError("Init param should be a dict.")
        return super_lineage_objs

    def filter_summary_lineage(self, condition=None):
        """
        Filter and sort lineage information based on the specified condition.

        See `ConditionType` and `ExpressionType` class for the rule of filtering
        and sorting. The filtering and sorting fields are defined in
        `FIELD_MAPPING` or prefixed with `metric/` or 'user_defined/'.

        If the condition is `None`, all model lineage information will be
        returned.

        Args:
            condition (Union[dict, None]): Filter and sort condition.
                Default: None.

        Returns:
            dict, filtered and sorted model lineage information.
        """
        def _filter(super_lineage_obj: SuperLineageObj):
            for condition_key, condition_value in condition.items():
                if ConditionParam.is_condition_type(condition_key):
                    continue
                if self._is_valid_field(condition_key):
                    raise LineageQuerierParamException(
                        'condition',
                        'The field {} not supported'.format(condition_key)
                    )

                value = super_lineage_obj.lineage_obj.get_value_by_key(condition_key)
                for exp_key, exp_value in condition_value.items():
                    if not ExpressionType.is_valid_exp(exp_key):
                        raise LineageQuerierParamException(
                            'condition',
                            'The expression {} not supported.'.format(exp_key)
                        )
                    if not ExpressionType.is_match(exp_key, exp_value, value):
                        return False
            return True

        if condition is None:
            condition = {}

        self._add_dataset_mark()
        super_lineage_objs = list(self._super_lineage_objs.values())
        super_lineage_objs.sort(key=lambda x: x.update_time, reverse=True)

        results = list(filter(_filter, super_lineage_objs))
        results = self._sorted_results(results, condition)

        offset_results = self._handle_limit_and_offset(condition, results)

        customized = self._organize_customized(offset_results)

        lineage_types = condition.get(ConditionParam.LINEAGE_TYPE.value)
        lineage_types = self._get_lineage_types(lineage_types)

        object_items = []
        for item in offset_results:
            lineage_object = dict()
            if LineageType.MODEL.value in lineage_types:
                lineage_object.update(item.lineage_obj.to_model_lineage_dict())
            if LineageType.DATASET.value in lineage_types:
                lineage_object.update(item.lineage_obj.to_dataset_lineage_dict())
            lineage_object.update({"added_info": item.added_info})
            object_items.append(lineage_object)

        lineage_info = {
            'customized': customized,
            'object': object_items,
            'count': len(results)
        }

        return lineage_info

    def _sorted_results(self, results, condition):
        """Get sorted results."""
        def _cmp(value1, value2):
            if value1 is None and value2 is None:
                cmp_result = 0
            elif value1 is None:
                cmp_result = -1
            elif value2 is None:
                cmp_result = 1
            else:
                try:
                    cmp_result = (value1 > value2) - (value1 < value2)
                except TypeError:
                    type1 = type(value1).__name__
                    type2 = type(value2).__name__
                    cmp_result = (type1 > type2) - (type1 < type2)
            return cmp_result

        def _cmp_added_info(obj1: SuperLineageObj, obj2: SuperLineageObj):
            value1 = obj1.added_info.get(sorted_name)
            value2 = obj2.added_info.get(sorted_name)
            return _cmp(value1, value2)

        def _cmp_super_lineage_obj(obj1: SuperLineageObj, obj2: SuperLineageObj):
            value1 = obj1.lineage_obj.get_value_by_key(sorted_name)
            value2 = obj2.lineage_obj.get_value_by_key(sorted_name)

            return _cmp(value1, value2)

        if ConditionParam.SORTED_NAME.value in condition:
            sorted_name = condition.get(ConditionParam.SORTED_NAME.value)
            sorted_type = condition.get(ConditionParam.SORTED_TYPE.value)
            reverse = sorted_type == 'descending'
            if sorted_name in ['tag']:
                results = sorted(
                    results, key=functools.cmp_to_key(_cmp_added_info), reverse=reverse
                )
                return results

            if self._is_valid_field(sorted_name):
                raise LineageQuerierParamException(
                    'condition',
                    'The sorted name {} not supported.'.format(sorted_name)
                )
            results = sorted(
                results, key=functools.cmp_to_key(_cmp_super_lineage_obj), reverse=reverse
            )
        return results

    def _organize_customized(self, offset_results):
        """Organize customized."""
        customized = dict()
        for offset_result in offset_results:
            for obj_name in ["metric", "user_defined"]:
                self._organize_customized_item(customized, offset_result.lineage_obj, obj_name)

        # If types contain numbers and string, it will be "mixed".
        # If types contain "int" and "float", it will be "float".
        for key, value in customized.items():
            types = value["type"]
            if len(types) == 1:
                customized[key]["type"] = list(types)[0]
            elif types.issubset(["int", "float"]):
                customized[key]["type"] = "float"
            else:
                customized[key]["type"] = "mixed"
        return customized

    def _organize_customized_item(self, customized, offset_result, obj_name):
        """Organize customized item."""
        obj = getattr(offset_result, obj_name)
        require = bool(obj_name == "metric")
        if obj and isinstance(obj, dict):
            for key, value in obj.items():
                label = f'{obj_name}/{key}'
                current_type = type(value).__name__
                if customized.get(label) is None:
                    customized[label] = dict()
                    customized[label]["label"] = label
                    # user defined info is not displayed by default
                    customized[label]["required"] = require
                    customized[label]["type"] = set()
                customized[label]["type"].add(current_type)

    def _get_lineage_types(self, lineage_type_param):
        """
        Get lineage types.

        Args:
            lineage_type_param (dict): A dict contains "in" or "eq".

        Returns:
            list, lineage type.

        """
        # lineage_type_param is None or an empty dict
        if not lineage_type_param:
            return enum_to_list(LineageType)

        if lineage_type_param.get("in") is not None:
            return lineage_type_param.get("in")

        return [lineage_type_param.get("eq")]

    def _is_valid_field(self, field_name):
        """
        Check if field name is valid.

        Args:
            field_name (str): Field name.

        Returns:
            bool, `True` if the field name is valid, else `False`.
        """
        return field_name not in FIELD_MAPPING \
               and not field_name.startswith(('metric/', 'user_defined/'))

    def _handle_limit_and_offset(self, condition, result):
        """
        Handling the condition of `limit` and `offset`.

        Args:
            condition (dict): Filter and sort condition.
            result (list[SuperLineageObj]): Filtered and sorted result.

        Returns:
            list[SuperLineageObj], paginated result.
        """
        offset = 0
        limit = 10
        if ConditionParam.OFFSET.value in condition:
            offset = condition.get(ConditionParam.OFFSET.value)
        if ConditionParam.LIMIT.value in condition:
            limit = condition.get(ConditionParam.LIMIT.value)
        if ConditionParam.OFFSET.value not in condition \
                and ConditionParam.LIMIT.value not in condition:
            offset_result = result
        else:
            offset_result = result[offset * limit: limit * (offset + 1)]
        return offset_result

    def _add_dataset_mark(self):
        """Add dataset mark into LineageObj."""
        # give a dataset mark for each dataset graph in lineage information
        marked_dataset_group = {'1': None}
        for super_lineage_obj in self._super_lineage_objs.values():
            lineage = super_lineage_obj.lineage_obj
            dataset_mark = '0'
            for dataset_graph_mark, marked_dataset_graph in marked_dataset_group.items():
                if marked_dataset_graph == lineage.dataset_graph:
                    dataset_mark = dataset_graph_mark
                    break
            # if no matched, add the new dataset graph into group
            if dataset_mark == '0':
                dataset_mark = str(int(max(marked_dataset_group.keys())) + 1)
                marked_dataset_group.update({
                    dataset_mark:
                        lineage.dataset_graph
                })
            lineage.dataset_mark = dataset_mark



# Copyright 2020-2021 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Explain job list encapsulator."""

from datetime import datetime

from mindinsight.explainer.encapsulator.explain_data_encap import ExplainDataEncap
from mindinsight.datavisual.common.exceptions import TrainJobNotExistError


class ExplainJobEncap(ExplainDataEncap):
    """Explain job list encapsulator."""

    DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
    DEFAULT_MIN_CONFIDENCE = 0.5

    def query_explain_jobs(self, offset, limit):
        """
        Query explain job list.

        Args:
            offset (int): Page offset.
            limit (int): Maximum number of items to be returned.

        Returns:
            tuple[int, list[Dict]], total number of jobs and job list.
        """
        total, dir_infos = self.job_manager.get_job_list(offset=offset, limit=limit)
        job_infos = [self._dir_2_info(dir_info) for dir_info in dir_infos]

        return total, job_infos

    def query_meta(self, train_id):
        """
        Query explain job meta-data.

        Args:
            train_id (str): Job ID.

        Returns:
            dict, the metadata.
        """
        job = self.job_manager.get_job(train_id)
        if job is None:
            raise TrainJobNotExistError(train_id)
        return self._job_2_meta(job)

    @classmethod
    def _dir_2_info(cls, dir_info):
        """Convert ExplainJob object to jsonable info object."""
        info = dict()
        info["train_id"] = dir_info["relative_path"]
        info["create_time"] = dir_info["create_time"].strftime(cls.DATETIME_FORMAT)
        info["update_time"] = dir_info["update_time"].strftime(cls.DATETIME_FORMAT)
        info["saliency_map"] = dir_info["saliency_map"]
        info["hierarchical_occlusion"] = dir_info["hierarchical_occlusion"]

        return info

    @classmethod
    def _job_2_info(cls, job):
        """Convert ExplainJob object to jsonable info object."""
        info = dict()
        info["train_id"] = job.train_id
        info["create_time"] = datetime.fromtimestamp(job.create_time)\
            .strftime(cls.DATETIME_FORMAT)
        info["update_time"] = datetime.fromtimestamp(job.update_time)\
            .strftime(cls.DATETIME_FORMAT)
        return info

    @classmethod
    def _job_2_meta(cls, job):
        """Convert ExplainJob's meta-data to jsonable info object."""
        info = cls._job_2_info(job)
        info["sample_count"] = job.sample_count
        info["classes"] = job.all_classes
        saliency_info = dict()
        if job.min_confidence is None:
            saliency_info["min_confidence"] = cls.DEFAULT_MIN_CONFIDENCE
        else:
            saliency_info["min_confidence"] = job.min_confidence
        saliency_info["explainers"] = list(job.explainers)
        saliency_info["metrics"] = list(job.metrics)
        info["saliency"] = saliency_info
        info["uncertainty"] = {"enabled": job.uncertainty_enabled}
        info["status"] = job.status
        return info



# Copyright 2019 Huawei Technologies Co., Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Handle custom error."""
from urllib.parse import quote
from werkzeug.exceptions import NotFound
from werkzeug.exceptions import MethodNotAllowed

from flask import request, jsonify

from mindinsight.datavisual.common.exceptions import RequestMethodNotAllowed
from mindinsight.datavisual.common.exceptions import RestfulApiNotExist
from mindinsight.datavisual.common.log import restful_logger as logger
from mindinsight.utils.exceptions import UnknownError
from mindinsight.utils.exceptions import FileSystemPermissionError


def handle_http_exception_error(ex):
    """Handle http exception error."""
    logger.warning('%r %r, detail: %r', request.method, quote(request.path), str(ex))
    if isinstance(ex, NotFound):
        error = RestfulApiNotExist()
    elif isinstance(ex, MethodNotAllowed):
        error = RequestMethodNotAllowed()
    else:
        logger.exception(ex)
        error = UnknownError('System error or http error.')
    res_body = {"error_code": error.error_code, "error_msg": error.message}
    return jsonify(res_body), error.http_code


def handle_mindinsight_error(ex):
    """Handle mindinsight error."""
    if int(ex.http_code) < 500:
        logger.warning('%r %r detail: %r', request.method, quote(request.path), ex.message)
    else:
        logger.error('%r %r detail: %r', request.method, quote(request.path), ex.message)
        logger.exception(ex)
    res_body = dict(error_code=ex.error_code, error_msg=ex.message)
    return jsonify(res_body), ex.http_code


def handle_unknown_error(ex):
    """Handle unknown error."""
    logger.error('%r %r detail: %r', request.method, quote(request.path), str(ex))
    logger.exception(ex)
    if isinstance(ex, PermissionError):
        error = FileSystemPermissionError('File System Permission Error')
    else:
        error = UnknownError('System error.')
    res_body = dict(error_code=error.error_code, error_msg=error.message)
    return jsonify(res_body), error.http_code
